
from typing import List
from ceruleo.transformation.functional.transformerstep import TransformerStep
import pandas as pd
from functools import reduce 


class Sum(TransformerStep):
    def transform(self, X: List[pd.DataFrame]):
        return reduce(lambda x, y: x.add(y, fill_value=0), X)


class Divide(TransformerStep):
    def transform(self, X: List[pd.DataFrame]):
        return reduce(lambda x, y: x.divide(y, fill_value=0), X)


from abc import abstractmethod

from signal import signal
from typing import Any, Callable, Union

import numpy as np



class AbstractSampleWeights:
    """The base class for the sample weight provider
    """
    def __call__(self, y, i: int, metadata):
        raise NotImplementedError


class NotWeighted(AbstractSampleWeights):
    """Simplest sample weight provvider

    Provide 1 as a sample weight for every sample
    """
    def __call__(self, y, i: int, metadata):
        return 1

"""The Sample Weight type is a callable with the following signature

    fun(y, i:int, metadata)

Given the target and the sample index `i` it returns the sample weight for sample `i`
"""
SampleWeight = Union[AbstractSampleWeights, Callable[[np.ndarray, int, Any], float]]


class RULInverseWeighted(AbstractSampleWeights):
    r"""Weight each sample inverse to the RUL

    ```math
    w_i = \frac{1}{RUL_{i} + 1} 
    ```
    """
    def __call__(self, y, i: int, metadata):
        return 1 / (y[i, 0] + 1)


class InverseToLengthWeighted(AbstractSampleWeights):
    """Weights samples according to the duration of the run-to-failure cycle they bolng

    All points in the run-to-cycle are weighted equally inverse to the cycle duration
    """
    def __call__(self, y, i: int, metadata):
        return 1 / y[0]


class ExponentialDecay(AbstractSampleWeights):
    def __init__(self, *, near_0_at:float):
        super().__init__()
        self.alpha = -(near_0_at)**2/np.log(0.000001)

    def __call__(self, y, i: int, metadata):
        return (1 + np.exp(-(y[i, 0]**2) / self.alpha))**2
        



from typing import Optional

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep


class TargetToClasses(TransformerStep):
    def __init__(self, bins):
        super().__init__()
        self.bins = bins

    def transform(self, X):
        X_new = pd.DataFrame(index=X.index)
        X_new["RUL_class"] = np.digitize(X.iloc[:, 0].values, self.bins, right=False)
        return X_new


class HealthPercentage(TransformerStep):
    def transform(self, X):
        return (X / X.iloc[0, 0]) * 100


class PicewiseRUL(TransformerStep):
    """Transform the RUL clipping the maximum value

    Parameters
    ----------
    max_life : float, optional
        Maximum threshold for clipping the RUL values, by default np.inf
    name : Optional[str], optional
        Name of the step, by default None    """
    def __init__(self, *, max_life: float = np.inf, name: Optional[str] = None):
        super().__init__(name=name)
        self.max_life = max_life

    def transform(self, X):
        """Clip the maximum value of the true RUL

        Parameters
        ----------
        X : np.array
            Vector with the true labels

        Returns
        -------
        np.array
            Cliiped RUL values
        """
        return np.clip(X, 0, self.max_life)


class PicewiseRULQuantile(PicewiseRUL):
    """Transform the RUL clipping the maximum value using a quantile value as threshold

    Parameters
    ----------
    quantile : float, optional
        Value between 0 and 1 to compute the RUL threshold
    name : Optional[str], optional
        Name of the step, by default None    """
    def __init__(self, quantile:float, name: Optional[str] = None):
        super().__init__(name)
        self.quantile = quantile

    def fit(self, X, y=None):
        """Clip the maximum value of the true RUL

        Parameters
        ----------
        X : np.array
            Vector with the true labels

        Returns
        -------
        np.array
            Cliiped RUL values
        """
        self.max_life = np.quantile(X, self.quantile)
        return self


class RULBinarizer(TransformerStep):
    def __init__(self, t:float, **kwargs):
        super().__init__(**kwargs)
        self.t = t       
    

    def transform(self, X):
        return (X < self.t).astype('int')




class RemoveGaps(TransformerStep):
    """Gaps larger than a treshold are removed

    Parameters
    ----------
    max_life : float, optional
        Maximum threshold for clipping the RUL values, by default np.inf
    name : Optional[str], optional
        Name of the step, by default None    """
    def __init__(self, *, threshold: float, name: Optional[str] = None):
        super().__init__(name=name)
        self.threshold = threshold

    def transform(self, x:pd.DataFrame):
        """Remove gaps in the RUL target

        Parameters
        ----------
        X : np.array
            Vector with the true labels

        Returns
        -------
        np.array
            Cliiped RUL values
        """

        y = np.abs(np.diff(np.squeeze(x)))
        y[y  <= self.threshold] = 0        
        new_x  = np.squeeze(x.values) + np.hstack((0, np.cumsum(y)))
        
        return pd.DataFrame(new_x - np.min(new_x), index=x.index, columns=x.columns)
        


import numpy as np 


def _strided_app(a, L, S):  # Window len = L, Stride len/stepsize = S
    """returns an array that is strided
    """
    nrows = ((a.size-L)//S)+1
    n = a.strides[0]
    r =  np.lib.stride_tricks.as_strided(
        a, 
        shape=(nrows, L), 
        strides=(S*n, n),
        writeable=False)
    last = r.shape[0]*r.shape[1]
    l = r.tolist()
    if last < len(a):
        l.append(a[last:])
    return l
    

def apply_rolling_data(values : np.ndarray, function, window, step=1):
    """Perform a rolling window analysis at the column `col` from `data`

    Given a dataframe `data` with time series, call `function` at
    sections of length `window` at the data of column `col`. Append
    the results to `data` at a new columns with name `label`.

    Parameters:
    
    data: 1-D Time series of data
    function: Function to be called to calculate the rolling window
        analysis, the function must receive as input an array or
        pandas series. Its output must be either a number or a pandas
        series
    window: length of the window to perform the analysis
    step: step to take between two consecutive windows

    Returns:
    -------
        data: Columns generated by the function applied

    """

    x = _strided_app(values, window, step)

    return np.vstack([function(np.array(b)) for b in x])
    



import gzip
import io
import logging
import os
import pickle
import shutil
import tarfile
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import gdown
import numpy as np
import pandas as pd
from ceruleo import CACHE_PATH, DATA_PATH
from ceruleo.dataset.ts_dataset import AbstractLivesDataset
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)



COMPRESSED_FILE = "phm_data_challenge_2018.tar.gz"
FOLDER = "phm_data_challenge_2018"


URL = "https://drive.google.com/uc?id=15Jx9Scq9FqpIGn8jbAQB_lcHSXvIoPzb"
OUTPUT = COMPRESSED_FILE


def download(path: Path):
    logger.info("Downloading dataset...")
    gdown.download(URL, str(path / OUTPUT), quiet=False)


def prepare_raw_dataset(path: Path):
    def track_progress(members):
        for member in tqdm(members, total=70):
            yield member

    path = path / "raw"
    path.mkdir(parents=True, exist_ok=True)
    if not (path / OUTPUT).resolve().is_file():
        download(path)
    logger.info("Decompressing  dataset...")
    with tarfile.open(path / OUTPUT, "r") as tarball:
        tarball.extractall(path=path, members=track_progress(tarball))
    shutil.move(str(path / "phm_data_challenge_2018" / "train"), str(path / "train"))
    shutil.move(str(path / "phm_data_challenge_2018" / "test"), str(path / "test"))
    shutil.rmtree(str(path / "phm_data_challenge_2018"))
    (path / OUTPUT).unlink()


class FailureType(Enum):
    """Failure types availables for the dataset.

    Possible values are

    ```py
    FailureType.FlowCoolPressureDroppedBelowLimit
    FailureType.FlowcoolPressureTooHighCheckFlowcoolPump
    FailureType.FlowcoolLeak
    ```

    """
    FlowCoolPressureDroppedBelowLimit = "FlowCool Pressure Dropped Below Limit"
    FlowcoolPressureTooHighCheckFlowcoolPump = (
        "Flowcool Pressure Too High Check Flowcool Pump"
    )
    FlowcoolLeak = "Flowcool leak"

    @staticmethod
    def that_starth_with(s: str):
        for f in FailureType:
            if s.startswith(f.value):
                return f
        return None


from typing import List, Optional, Union


def merge_data_with_faults(
    data_file: Union[str, Path], fault_data_file: Union[str, Path]
) -> pd.DataFrame:
    """Merge the raw sensor data with the fault information

    Parameters:

        data_file: Path where the raw sensor data is located
        fault_data_file: Path where the fault information is located

    Returns:

        df: Dataframe indexed by time with the raw sensors and faults
            The dataframe contains also a fault_number column
    """
    data = pd.read_csv(data_file).set_index("time")

    fault_data = (
        pd.read_csv(fault_data_file).drop_duplicates(subset=["time"]).set_index("time")
    )
    fault_data["fault_number"] = range(fault_data.shape[0])
    return pd.merge_asof(data, fault_data, on="time", direction="forward").set_index(
        "time"
    )


def prepare_dataset(dataset_path: Path):
    (dataset_path / "processed" / "lives").mkdir(exist_ok=True, parents=True)

    files = list(Path(dataset_path / "raw" / "train").resolve().glob("*.csv"))
    faults_files = list(
        Path(dataset_path / "raw" / "train" / "train_faults").resolve().glob("*.csv")
    )
    files = {file.stem[0:6]: file for file in files}
    faults_files = {file.stem[0:6]: file for file in faults_files}
    dataset_data = []
    for filename in tqdm(faults_files.keys(), "Processing files"):
        tool = filename[0:6]
        data_file = files[tool]
        logger.info(f"Loading data file {files[tool]}")
        fault_data_file = faults_files[filename]
        data = merge_data_with_faults(data_file, fault_data_file)
        for life_index, life_data in data.groupby("fault_number"):
            if life_data.shape[0] == 0:
                continue
            failure_type = FailureType.that_starth_with(life_data["fault_name"].iloc[0])
            output_filename = (
                f"Life_{int(life_index)}_{tool}_{failure_type.name}.pkl.gzip"
            )
            dataset_data.append(
                (tool, life_data.shape[0], failure_type.value, output_filename)
            )
            life = life_data.copy()
            life["RUL"] = np.arange(life.shape[0] - 1, -1, -1)
            with gzip.open(
                dataset_path / "processed" / "lives" / output_filename, "wb"
            ) as file:
                pickle.dump(life_data, file)

    df = pd.DataFrame(
        dataset_data, columns=["Tool", "Number of samples", "Failure Type", "Filename"]
    )
    df.to_csv(dataset_path / "processed" / "lives" / "lives_db.csv")


class PHMDataset2018(AbstractLivesDataset):
    """PHM 2018 Dataset

    The 2018 PHM dataset is a public dataset released by Seagate which contains the execution of 20 different 
    ion milling machines. They distinguish three different failure causes and provide 22 features, 
    including user-defined variables and sensors. 

    Three faults are present in the dataset

    * Fault mode 1 occurs when flow-cool pressure drops.
    * Fault mode 2 occurs when flow-cool pressure becomes too high.
    * Fault mode 3 represents flow-cool leakage.

    [Dataset reference](https://phmsociety.org/conference/annual-conference-of-the-phm-society/annual-conference-of-the-prognostics-and-health-management-society-2018-b/phm-data-challenge-6/)

    Example:
 
    ```py
    dataset = PHMDataset2018(
       failure_types=FailureType.FlowCoolPressureDroppedBelowLimit,
        tools=['01_M02']
    )
    ```


    
    Parameters:

        failure_types: List of failure types
        tools: List of tools
        path: Path where the dataset is located
    """
    def __init__(
        self,
        failure_types: Union[FailureType, List[FailureType]] = [l for l in FailureType],
        tools: Union[str, List[str]] = "all",
        path: Path = DATA_PATH,
    ):
        super().__init__()
        if not isinstance(failure_types, list):
            failure_types = [failure_types]
        self.failure_types = failure_types

        if isinstance(tools, str) and tools != "all":
            tools = [tools]
        self.tools = tools

        self.path = path
        self.dataset_path = path / FOLDER

        self.procesed_path = self.dataset_path / "processed" / "lives"
        self.lives_table_filename = self.procesed_path / "lives_db.csv"
        if not self.lives_table_filename.is_file():
            if not (self.dataset_path / "raw" / "train").is_dir():
                prepare_raw_dataset(self.dataset_path)
            prepare_dataset(self.dataset_path)

        self.lives = pd.read_csv(self.lives_table_filename)
        if tools != "all":
            self.lives = self.lives[self.lives["Tool"].isin(self.tools)]
        self.lives = self.lives[
            self.lives["Failure Type"].isin([a.value for a in self.failure_types])
        ]
        valid = []
        for i, (j, r) in enumerate(self.lives.iterrows()):
            df = self._load_life(r["Filename"])
            if df.shape[0] > 1200:
                valid.append(i)
        self.lives = self.lives.iloc[valid, :]

    @property
    def n_time_series(self) -> int:
        return self.lives.shape[0]

    def _load_life(self, filename: str) -> pd.DataFrame:
        with gzip.open(self.procesed_path / filename, "rb") as file:
            df = pickle.load(file)
        return df

    def get_time_series(self, i: int) -> pd.DataFrame:
        df = self._load_life(self.lives.iloc[i]["Filename"])
        df.index = pd.to_timedelta(df.index, unit='s')
        df = df[df['FIXTURESHUTTERPOSITION'] == 1]
        df["RUL"] = np.arange(df.shape[0] - 1, -1, -1)

        return df

    @property
    def rul_column(self) -> str:
        return "RUL"



import logging
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractTimeSeriesDataset

logger = logging.getLogger(__name__)


def sample_rate(ds: AbstractTimeSeriesDataset, unit: Optional[str] = "s") -> np.ndarray:
    """Obtain an array of time difference between two consecutive samples

    If the index it's a timestamp, the time difference will be converted to the provided
    unit

    Parameters:
        ds: The dataset
        unit: Unit to convert the timestamps differences

    Returns:
    
        sample_rates: np.ndarray

    """
    time_diff = []
    for life in ds:
        diff = np.diff(life.index.values)
        if pd.api.types.is_timedelta64_ns_dtype(diff.dtype):
            if unit is None:
                unit = "s"
            diff = diff / np.timedelta64(1, unit)
        time_diff.extend(diff)
    return np.array(time_diff)


def sample_rate_summary(
    ds: AbstractTimeSeriesDataset, unit: Optional[str] = "s"
) -> pd.DataFrame:
    """Obtain the main and standard deviation of the sample rate of the dataset

    Parameters:

        ds: The dataset
        unit: Unit to convert the time differences

    Returns:
        df: Dataframe with the following columns        
            Mean sample rate,  Std sample rate, Mode sample rate
    """
    sr = sample_rate(ds, unit)
    return pd.DataFrame(
        {
            "Mean sample rate": np.mean(sr),
            "Std sample rate": np.std(sr),
            "Mode sample rate": pd.Series(sr).mode().values[0],
        },
        index=["Dataset"],
    )


"""
In some cases It's useful to cast the datatype of subset of columns

"""
import logging
from typing import Optional

import pandas as pd
from ceruleo.transformation import TransformerStep

logger = logging.getLogger(__name__)


class CastTo(TransformerStep):
    """Cast to a given datatype

    Example:

        step = CastTo(type='float32')

    Parameters:

        type: Type name to convert to
        name: Name of the step

    """
    def __init__(self, *, type:str, name:Optional[str]=None):
        super().__init__(name=name)
        self.type = type


    def transform(self, X: pd.DataFrame):
        return X.astype(self.type)



"""
Attention meachanism
"""

import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.layers import (Add, Conv1D, Dense, Dropout, Lambda,
                                     Permute)


class Attention(tf.keras.Model):
    """
    Temporal pattern attention for multivariate time series forecasting
    Shun-Yao Shih, Fan-Keng Sun & Hung-yi Lee

    They propose using a set of filters to extract time-invariant temporal patterns,
    similar to transforming time series data into its “frequency domain”.
    Then they propose a novel attention mechanism to select relevant time series,
    and use its frequency domain information for multivariate forecasting.
    """

    def __init__(self, number_of_filters, attention_size, dropout=0.1):
        super(Attention, self).__init__()
        self.filter_size = 1
        self.number_of_filters = number_of_filters
        self.attention_size = attention_size
        self.dropout = dropout

        self.permute = Permute((2, 1))
        self.expand = Lambda(lambda x: K.expand_dims(x))
        self.dropout = Dropout(self.dropout)
        self.squeeze = Lambda(lambda x: K.squeeze(x, 1))
        self.prev_states = Lambda(lambda x: x[:, :-1, :])
        self.last_state = Lambda(lambda x: x[:, -1, :])
        self.permute1 = Permute((2, 1))

    def build(self, input_shape):
        self.W = self.add_weight(name="att_weight",
                                 shape=(self.number_of_filters,
                                        input_shape[2]),
                                 trainable=True)

        self.conv_layer = Conv1D(self.number_of_filters,
                                 self.attention_size,
                                 padding='same')
        self.denseHt = Dense(self.number_of_filters, use_bias=False)
        self.denseVt = Dense(self.number_of_filters, use_bias=False)

        super(Attention, self).build(input_shape)

    def call(self, x):

        prev_states = self.prev_states(x)
        prev_states = self.permute(prev_states)
        prev_states = self.conv_layer(prev_states)
        prev_states = self.dropout(prev_states)
        last_state = self.last_state(x)
        score = K.sigmoid(K.batch_dot(
            prev_states, K.dot(last_state, K.transpose(self.W))))

        score = K.expand_dims(score, axis=2)
        score = K.repeat_elements(score, rep=prev_states.shape[2], axis=2)
        vt = K.sum(tf.math.multiply(score, prev_states), axis=2)

        vt = Add()([self.denseHt(vt), self.denseVt(last_state)])

        return vt



# TAKEN FROM https://github.com/uchidalab/time_series_augmentation

from typing import List, Optional
import numpy as np
from tqdm import tqdm


def jitter(x, sigma=0.03):
    # https://arxiv.org/pdf/1706.00527.pdf
    return x + np.random.normal(loc=0.0, scale=sigma, size=x.shape)


def scaling(x, sigma=0.1):
    # https://arxiv.org/pdf/1706.00527.pdf
    factor = np.random.normal(loc=1.0, scale=sigma, size=(x.shape[0], x.shape[2]))
    return np.multiply(x, factor[:, np.newaxis, :])


def rotation(x):
    flip = np.random.choice([-1, 1], size=(x.shape[0], x.shape[2]))
    rotate_axis = np.arange(x.shape[2])
    np.random.shuffle(rotate_axis)
    return flip[:, np.newaxis, :] * x[:, :, rotate_axis]


def permutation(x, max_segments=5, seg_mode="equal"):
    orig_steps = np.arange(x.shape[1])

    num_segs = np.random.randint(1, max_segments, size=(x.shape[0]))

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        if num_segs[i] > 1:
            if seg_mode == "random":
                split_points = np.random.choice(
                    x.shape[1] - 2, num_segs[i] - 1, replace=False
                )
                split_points.sort()
                splits = np.split(orig_steps, split_points)
            else:
                splits = np.array_split(orig_steps, num_segs[i])
            warp = np.concatenate(np.random.permutation(splits)).ravel()
            ret[i] = pat[warp]
        else:
            ret[i] = pat
    return ret


def magnitude_warp(x, sigma=0.2, knot=4):
    from scipy.interpolate import CubicSpline

    orig_steps = np.arange(x.shape[1])

    random_warps = np.random.normal(
        loc=1.0, scale=sigma, size=(x.shape[0], knot + 2, x.shape[2])
    )
    warp_steps = (
        np.ones((x.shape[2], 1)) * (np.linspace(0, x.shape[1] - 1.0, num=knot + 2))
    ).T
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        warper = np.array(
            [
                CubicSpline(warp_steps[:, dim], random_warps[i, :, dim])(orig_steps)
                for dim in range(x.shape[2])
            ]
        ).T
        ret[i] = pat * warper

    return ret


def time_warp(x, sigma=0.2, knot=4):
    from scipy.interpolate import CubicSpline

    orig_steps = np.arange(x.shape[1])

    random_warps = np.random.normal(
        loc=1.0, scale=sigma, size=(x.shape[0], knot + 2, x.shape[2])
    )
    warp_steps = (
        np.ones((x.shape[2], 1)) * (np.linspace(0, x.shape[1] - 1.0, num=knot + 2))
    ).T

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            time_warp = CubicSpline(
                warp_steps[:, dim], warp_steps[:, dim] * random_warps[i, :, dim]
            )(orig_steps)
            scale = (x.shape[1] - 1) / time_warp[-1]
            ret[i, :, dim] = np.interp(
                orig_steps, np.clip(scale * time_warp, 0, x.shape[1] - 1), pat[:, dim]
            ).T
    return ret


def window_slice(x, reduce_ratio=0.9):
    # https://halshs.archives-ouvertes.fr/halshs-01357973/document
    target_len = np.ceil(reduce_ratio * x.shape[1]).astype(int)
    if target_len >= x.shape[1]:
        return x
    starts = np.random.randint(
        low=0, high=x.shape[1] - target_len, size=(x.shape[0])
    ).astype(int)
    ends = (target_len + starts).astype(int)

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            ret[i, :, dim] = np.interp(
                np.linspace(0, target_len, num=x.shape[1]),
                np.arange(target_len),
                pat[starts[i] : ends[i], dim],
            ).T
    return ret


def window_warp(x, window_ratio=0.1, scales=[0.5, 2.0]):
    # https://halshs.archives-ouvertes.fr/halshs-01357973/document
    warp_scales = np.random.choice(scales, x.shape[0])
    warp_size = np.ceil(window_ratio * x.shape[1]).astype(int)
    window_steps = np.arange(warp_size)

    window_starts = np.random.randint(
        low=1, high=x.shape[1] - warp_size - 1, size=(x.shape[0])
    ).astype(int)
    window_ends = (window_starts + warp_size).astype(int)

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            start_seg = pat[: window_starts[i], dim]
            window_seg = np.interp(
                np.linspace(0, warp_size - 1, num=int(warp_size * warp_scales[i])),
                window_steps,
                pat[window_starts[i] : window_ends[i], dim],
            )
            end_seg = pat[window_ends[i] :, dim]
            warped = np.concatenate((start_seg, window_seg, end_seg))
            ret[i, :, dim] = np.interp(
                np.arange(x.shape[1]),
                np.linspace(0, x.shape[1] - 1.0, num=warped.size),
                warped,
            ).T
    return ret


RETURN_VALUE = 0
RETURN_PATH = 1
RETURN_ALL = -1


# Core DTW
def _traceback(DTW, slope_constraint):
    i, j = np.array(DTW.shape) - 1
    p, q = [i-1], [j-1]
    
    if slope_constraint == "asymmetric":
        while (i > 1):
            tb = np.argmin((DTW[i-1, j], DTW[i-1, j-1], DTW[i-1, j-2]))

            if (tb == 0):
                i = i - 1
            elif (tb == 1):
                i = i - 1
                j = j - 1
            elif (tb == 2):
                i = i - 1
                j = j - 2

            p.insert(0, i-1)
            q.insert(0, j-1)
    elif slope_constraint == "symmetric":
        while (i > 1 or j > 1):
            tb = np.argmin((DTW[i-1, j-1], DTW[i-1, j], DTW[i, j-1]))

            if (tb == 0):
                i = i - 1
                j = j - 1
            elif (tb == 1):
                i = i - 1
            elif (tb == 2):
                j = j - 1

            p.insert(0, i-1)
            q.insert(0, j-1)
        
        
    return (np.array(p), np.array(q))


def dtw(prototype, sample, return_flag = RETURN_VALUE, slope_constraint="asymmetric", window=None):
    """ Computes the DTW of two sequences.
    :param prototype: np array [0..b]
    :param sample: np array [0..t]
    :param extended: bool
    """
    p = prototype.shape[0]
    assert p != 0, "Prototype empty!"
    s = sample.shape[0]
    assert s != 0, "Sample empty!"
    
    if window is None:
        window = s
    
    cost = np.full((p, s), np.inf)
    for i in range(p):
        start = max(0, i-window)
        end = min(s, i+window)+1
        cost[i,start:end]=np.linalg.norm(sample[start:end] - prototype[i], axis=1)

    DTW = _cummulative_matrix(cost, slope_constraint, window)
        
    if return_flag == RETURN_ALL:
        return DTW[-1,-1], cost, DTW[1:,1:], _traceback(DTW, slope_constraint)
    elif return_flag == RETURN_PATH:
        return _traceback(DTW, slope_constraint)
    else:
        return DTW[-1,-1]

def spawner(x, labels, sigma=0.05, verbose=-1):
    # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6983028/
    # use verbose=-1 to turn off warnings
    # use verbose=1 to print out figures

    

    random_points = np.random.randint(low=1, high=x.shape[1] - 1, size=x.shape[0])
    window = np.ceil(x.shape[1] / 10.0).astype(int)
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    ret = np.zeros_like(x)
    for i, pat in enumerate(tqdm(x)):
        # guarentees that same one isnt selected
        choices = np.delete(np.arange(x.shape[0]), i)
        # remove ones of different classes
        choices = np.where(l[choices] == l[i])[0]
        if choices.size > 0:
            random_sample = x[np.random.choice(choices)]
            # SPAWNER splits the path into two randomly
            path1 = dtw.dtw(
                pat[: random_points[i]],
                random_sample[: random_points[i]],
                dtw.RETURN_PATH,
                slope_constraint="symmetric",
                window=window,
            )
            path2 = dtw.dtw(
                pat[random_points[i] :],
                random_sample[random_points[i] :],
                dtw.RETURN_PATH,
                slope_constraint="symmetric",
                window=window,
            )
            combined = np.concatenate(
                (np.vstack(path1), np.vstack(path2 + random_points[i])), axis=1
            )
            if verbose:
                print(random_points[i])
                dtw_value, cost, DTW_map, path = dtw.dtw(
                    pat,
                    random_sample,
                    return_flag=dtw.RETURN_ALL,
                    slope_constraint=slope_constraint,
                    window=window,
                )
                dtw.draw_graph1d(cost, DTW_map, path, pat, random_sample)
                dtw.draw_graph1d(cost, DTW_map, combined, pat, random_sample)
            mean = np.mean([pat[combined[0]], random_sample[combined[1]]], axis=0)
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(
                    orig_steps,
                    np.linspace(0, x.shape[1] - 1.0, num=mean.shape[0]),
                    mean[:, dim],
                ).T
        else:
            if verbose > -1:
                print(
                    "There is only one pattern of class %d, skipping pattern average"
                    % l[i]
                )
            ret[i, :] = pat
    return jitter(ret, sigma=sigma)


def wdba(
    x, labels, batch_size=6, slope_constraint="symmetric", use_window=True, verbose=0
):
    # https://ieeexplore.ieee.org/document/8215569
    # use verbose = -1 to turn off warnings
    # slope_constraint is for DTW. "symmetric" or "asymmetric"



    if use_window:
        window = np.ceil(x.shape[1] / 10.0).astype(int)
    else:
        window = None
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    ret = np.zeros_like(x)
    for i in tqdm(range(ret.shape[0])):
        # get the same class as i
        choices = np.where(l == l[i])[0]
        if choices.size > 0:
            # pick random intra-class pattern
            k = min(choices.size, batch_size)
            random_prototypes = x[np.random.choice(choices, k, replace=False)]

            # calculate dtw between all
            dtw_matrix = np.zeros((k, k))
            for p, prototype in enumerate(random_prototypes):
                for s, sample in enumerate(random_prototypes):
                    if p == s:
                        dtw_matrix[p, s] = 0.0
                    else:
                        dtw_matrix[p, s] = dtw.dtw(
                            prototype,
                            sample,
                            dtw.RETURN_VALUE,
                            slope_constraint=slope_constraint,
                            window=window,
                        )

            # get medoid
            medoid_id = np.argsort(np.sum(dtw_matrix, axis=1))[0]
            nearest_order = np.argsort(dtw_matrix[medoid_id])
            medoid_pattern = random_prototypes[medoid_id]

            # start weighted DBA
            average_pattern = np.zeros_like(medoid_pattern)
            weighted_sums = np.zeros((medoid_pattern.shape[0]))
            for nid in nearest_order:
                if nid == medoid_id or dtw_matrix[medoid_id, nearest_order[1]] == 0.0:
                    average_pattern += medoid_pattern
                    weighted_sums += np.ones_like(weighted_sums)
                else:
                    path = dtw.dtw(
                        medoid_pattern,
                        random_prototypes[nid],
                        dtw.RETURN_PATH,
                        slope_constraint=slope_constraint,
                        window=window,
                    )
                    dtw_value = dtw_matrix[medoid_id, nid]
                    warped = random_prototypes[nid, path[1]]
                    weight = np.exp(
                        np.log(0.5)
                        * dtw_value
                        / dtw_matrix[medoid_id, nearest_order[1]]
                    )
                    average_pattern[path[0]] += weight * warped
                    weighted_sums[path[0]] += weight

            ret[i, :] = average_pattern / weighted_sums[:, np.newaxis]
        else:
            if verbose > -1:
                print(
                    "There is only one pattern of class %d, skipping pattern average"
                    % l[i]
                )
            ret[i, :] = x[i]
    return ret


# Proposed


def random_guided_warp(
    x,
    labels,
    slope_constraint="symmetric",
    use_window=True,
    dtw_type="normal",
    verbose=0,
):
    # use verbose = -1 to turn off warnings
    # slope_constraint is for DTW. "symmetric" or "asymmetric"
    # dtw_type is for shapeDTW or DTW. "normal" or "shape"



    if use_window:
        window = np.ceil(x.shape[1] / 10.0).astype(int)
    else:
        window = None
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    ret = np.zeros_like(x)
    for i, pat in enumerate(tqdm(x)):
        # guarentees that same one isnt selected
        choices = np.delete(np.arange(x.shape[0]), i)
        # remove ones of different classes
        choices = np.where(l[choices] == l[i])[0]
        if choices.size > 0:
            # pick random intra-class pattern
            random_prototype = x[np.random.choice(choices)]

            if dtw_type == "shape":
                path = dtw.shape_dtw(
                    random_prototype,
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )
            else:
                path = dtw.dtw(
                    random_prototype,
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )

            # Time warp
            warped = pat[path[1]]
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(
                    orig_steps,
                    np.linspace(0, x.shape[1] - 1.0, num=warped.shape[0]),
                    warped[:, dim],
                ).T
        else:
            if verbose > -1:
                print(
                    "There is only one pattern of class %d, skipping timewarping" % l[i]
                )
            ret[i, :] = pat
    return ret


def random_guided_warp_shape(x, labels, slope_constraint="symmetric", use_window=True):
    return random_guided_warp(x, labels, slope_constraint, use_window, dtw_type="shape")


def discriminative_guided_warp(
    x,
    labels,
    batch_size=6,
    slope_constraint="symmetric",
    use_window=True,
    dtw_type="normal",
    use_variable_slice=True,
    verbose=0,
):
    # use verbose = -1 to turn off warnings
    # slope_constraint is for DTW. "symmetric" or "asymmetric"
    # dtw_type is for shapeDTW or DTW. "normal" or "shape"

    if use_window:
        window = np.ceil(x.shape[1] / 10.0).astype(int)
    else:
        window = None
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    positive_batch = np.ceil(batch_size / 2).astype(int)
    negative_batch = np.floor(batch_size / 2).astype(int)

    ret = np.zeros_like(x)
    warp_amount = np.zeros(x.shape[0])
    for i, pat in enumerate(tqdm(x)):
        # guarentees that same one isnt selected
        choices = np.delete(np.arange(x.shape[0]), i)

        # remove ones of different classes
        positive = np.where(l[choices] == l[i])[0]
        negative = np.where(l[choices] != l[i])[0]

        if positive.size > 0 and negative.size > 0:
            pos_k = min(positive.size, positive_batch)
            neg_k = min(negative.size, negative_batch)
            positive_prototypes = x[np.random.choice(positive, pos_k, replace=False)]
            negative_prototypes = x[np.random.choice(negative, neg_k, replace=False)]

            # vector embedding and nearest prototype in one
            pos_aves = np.zeros((pos_k))
            neg_aves = np.zeros((pos_k))
            if dtw_type == "shape":
                for p, pos_prot in enumerate(positive_prototypes):
                    for ps, pos_samp in enumerate(positive_prototypes):
                        if p != ps:
                            pos_aves[p] += (1.0 / (pos_k - 1.0)) * dtw.shape_dtw(
                                pos_prot,
                                pos_samp,
                                dtw.RETURN_VALUE,
                                slope_constraint=slope_constraint,
                                window=window,
                            )
                    for ns, neg_samp in enumerate(negative_prototypes):
                        neg_aves[p] += (1.0 / neg_k) * dtw.shape_dtw(
                            pos_prot,
                            neg_samp,
                            dtw.RETURN_VALUE,
                            slope_constraint=slope_constraint,
                            window=window,
                        )
                selected_id = np.argmax(neg_aves - pos_aves)
                path = dtw.shape_dtw(
                    positive_prototypes[selected_id],
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )
            else:
                for p, pos_prot in enumerate(positive_prototypes):
                    for ps, pos_samp in enumerate(positive_prototypes):
                        if p != ps:
                            pos_aves[p] += (1.0 / (pos_k - 1.0)) * dtw.dtw(
                                pos_prot,
                                pos_samp,
                                dtw.RETURN_VALUE,
                                slope_constraint=slope_constraint,
                                window=window,
                            )
                    for ns, neg_samp in enumerate(negative_prototypes):
                        neg_aves[p] += (1.0 / neg_k) * dtw.dtw(
                            pos_prot,
                            neg_samp,
                            dtw.RETURN_VALUE,
                            slope_constraint=slope_constraint,
                            window=window,
                        )
                selected_id = np.argmax(neg_aves - pos_aves)
                path = dtw.dtw(
                    positive_prototypes[selected_id],
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )

            # Time warp
            warped = pat[path[1]]
            warp_path_interp = np.interp(
                orig_steps,
                np.linspace(0, x.shape[1] - 1.0, num=warped.shape[0]),
                path[1],
            )
            warp_amount[i] = np.sum(np.abs(orig_steps - warp_path_interp))
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(
                    orig_steps,
                    np.linspace(0, x.shape[1] - 1.0, num=warped.shape[0]),
                    warped[:, dim],
                ).T
        else:
            if verbose > -1:
                print("There is only one pattern of class %d" % l[i])
            ret[i, :] = pat
            warp_amount[i] = 0.0
    if use_variable_slice:
        max_warp = np.max(warp_amount)
        if max_warp == 0:
            # unchanged
            ret = window_slice(ret, reduce_ratio=0.9)
        else:
            for i, pat in enumerate(ret):
                # Variable Sllicing
                ret[i] = window_slice(
                    pat[np.newaxis, :, :],
                    reduce_ratio=0.9 + 0.1 * warp_amount[i] / max_warp,
                )[0]
    return ret


def discriminative_guided_warp_shape(
    x, labels, batch_size=6, slope_constraint="symmetric", use_window=True
):
    return discriminative_guided_warp(
        x, labels, batch_size, slope_constraint, use_window, dtw_type="shape"
    )


functions = {
    "jitter": lambda x, y: jitter(x),
    "scaling": lambda x, y: scaling(x),
    "rotation": lambda x, y: rotation(x),
    "permutation": lambda x, y: permutation(x),
    "randompermutation": lambda x, y: permutation(x, seg_mode="random"),
    #"magwarp": lambda x, y: magnitude_warp(x),
    #"timewarp": lambda x, y: time_warp(x),
    #"windowslice": lambda x, y: window_slice(x),
    #"windowwarp": lambda x, y: window_warp(x),
    #"spawner": lambda x, y: spawner(x, y),
  #  "dtwwarp": random_guided_warp,
  #  "shapedtwwarp": random_guided_warp_shape,
    #"wdba": wdba,
  #  "discdtw": discriminative_guided_warp,
  #  "discsdtw": discriminative_guided_warp_shape,
}


def augment_dataset(x, y, N: int, augmentations: Optional[List[str]] = None):
    x_aug = x
    y_aug = y

    if augmentations is None:
        augmentations = sorted(list(augmentations.keys()))
    for n in range(N):
        x_temp = x
        for to_do in augmentations:
            x_temp = functions[to_do](x, y)
        x_aug = np.append(x_aug, x_temp, axis=0)
        y_aug = np.append(y_aug, y, axis=0)
    return x_aug, y_aug

def augment_sample( x, y):
    x_aug = x
    y_aug = y
    augmentations = None
    if augmentations is None:
        augmentations = sorted(list(functions.keys()))
    
    for to_do in augmentations:
        x_aug = functions[to_do](x_aug, y)
    
    return x_aug, y_aug


"""
Activation functions
"""

from tensorflow.keras.layers import Activation, multiply


def gated_activation(x):
    # Used in PixelCNN and WaveNet
    tanh = Activation('tanh')(x)
    sigmoid = Activation('sigmoid')(x)
    return multiply([tanh, sigmoid])




from collections.abc import Iterable
from re import S
from typing import Any, List, Tuple, Union

import numpy as np
import pandas as pd
import tensorflow as tf
from numpy.lib.arraysetops import isin
from tqdm.auto import tqdm
from abc import abstractmethod, abstractproperty

class DatasetIterator:
    def __init__(self, dataset):
        self.dataset = dataset
        self.i = 0

    def __next__(self):
        if self.i == self.dataset.n_time_series:
            raise StopIteration
        a = self.dataset[self.i]
        self.i += 1
        return a


class AbstractTimeSeriesDataset:
    def __init__(self):
        self._common_features = None
        self._durations = None

    def __iter__(self):
        return DatasetIterator(self)

    @property
    def n_time_series(self) -> int:
        raise NotImplementedError

    def get_time_series(self, i: int) -> pd.DataFrame:
        """

        Returns:
            df: DataFrame with the data of the life i
        """
        raise NotImplementedError

    def number_of_samples_of_time_series(self, i: int) -> int:
        return self[i].shape[0]

    def number_of_samples(self) -> List[int]:
        return [
            self.number_of_samples_of_time_series(i) for i in tqdm(range(len(self)))
        ]

    def duration(self, life: pd.DataFrame) -> float:
        """Obtain the duration of the time-series

        Parameters:
            i: Index of the life

        Returns:
            duration: Duration of the life
        """
        v = life.index
        return v.max() - v.min()

    def durations(self, show_progress: bool = False) -> List[float]:
        """Obtain the length of each life

        Return:

            durations: List of durations
        """
        if self._durations is None:
            if show_progress:
                iterator = tqdm(self)
            else:
                iterator = self
            self._durations = [self.duration(life) for life in iterator]
            # [self.rul_column].iloc[0]
        return self._durations

    def __call__(self, i):
        return self[i]

    def get_features_of_life(self, i:int ) -> pd.DataFrame:
        return self[i]

    def __getitem__(self, i: Union[int, Iterable]):
        """Obtain a time-series or an splice of the dataset using a FoldedDataset

        Parameters:

            i: If the parameter is an in it will return a pd.DataFrame with the i-th time-series.
                If the parameter is a list of int it will return a FoldedDataset with the
                time-series whose id are present in the list

        Raises:
            ValueError: When the list does not contain integer parameters

        Returns:
            pd.DataFrame: the i-th time-series
            FoldedDataset: The dataset with the lives specified by the list
        """
        if isinstance(i, slice):
            i = range(
                0 if i.start is None else i.start,
                len(self) if i.stop is None else i.stop,
                1 if i.step is None else i.step,
            )
        if isinstance(i, tf.Tensor):
            return self.get_time_series(i.ref())
        if isinstance(i, Iterable):
            if not all(isinstance(item, (int, np.integer)) for item in i):
                raise ValueError("Invalid iterable index passed")

            return FoldedDataset(self, i)
        else:
            df = self.get_time_series(i)
            return df

    @property
    def shape(self) -> Tuple[int, int]:
        return (self.n_time_series, 1)

    def __len__(self):
        """
        Return:

            n: The number of time-series in the dataset

        """
        return self.n_time_series

    def to_pandas(
        self,
        proportion_of_lives: float = 1.0,
        subsample_proportion: float = 1.0,
        show_progress: bool = False,
    ) -> pd.DataFrame:
        """
        Create a dataset with the time-series concatenated

        Parameters:

            proportion_of_lives: Proportion of lives to use.

        Returns:

            df: a DataFrame with all the lives concatenated
        """
        if show_progress:
            bar = tqdm
        else:
            bar = lambda x: x
        df = []

        features = list(
            self._compute_common_features(
                proportion_of_lives=proportion_of_lives, show_progress=show_progress
            )
        )

        for i in bar(range(self.n_time_series)):

            if proportion_of_lives < 1.0 and np.random.rand() > proportion_of_lives:
                continue

            current_life = self[i].loc[:, features]
            if subsample_proportion < 1.0:
                indices = range(
                    0,
                    current_life.shape[0],
                    int(current_life.shape[0] * subsample_proportion),
                )
                current_life = current_life.iloc[indices, :]
            df.append(current_life)
        return pd.concat(df)

    def _compute_common_features(
        self, proportion_of_lives: float = 1.0, show_progress: bool = False
    ) -> List[str]:
        common_features = []
        if show_progress:
            bar = tqdm
        else:
            bar = lambda x: x
        for i in bar(range(self.n_time_series)):
            if proportion_of_lives < 1.0 and np.random.rand() > proportion_of_lives:
                continue
            life = self.get_features_of_life(i)
            common_features.append(set(life.columns.values))
        return sorted(list(common_features[0].intersection(*common_features)))

    def common_features(
        self, show_progress: bool = False, proportion_of_lives: float = 1.0
    ) -> List[str]:
        if self._common_features is None:
            self._common_features = self._compute_common_features(
                proportion_of_lives, show_progress=show_progress
            )
        return self._common_features

    def map(self, transformer, cache_size: int = None):
        from ceruleo.dataset.transformed import TransformedDataset

        return TransformedDataset(self, transformer, cache_size=cache_size)

    def numeric_features(self, show_progress: bool = False) -> List[str]:
        """Obtain the list of the common numeric features in the dataset

        Parameters:

            show_progress : Whether to show progress when computing the common features, by default False

        Returns:

            l: List of columns
        """

        features = self.common_features(show_progress=show_progress)
        df = self.get_features_of_life(0)
        return list(
            df.loc[:, features]
            .select_dtypes(include=[np.number], exclude=["datetime", "timedelta"])
            .columns.values
        )

    def categorical_features(self, show_progress: bool = False) -> List[str]:
        features = self.common_features(show_progress=show_progress)
        df = self.get_time_series(0)
        return list(
            df.loc[:, features]
            .select_dtypes(exclude=[np.number, "datetime", "timedelta"])
            .columns.values
        )


class FoldedDataset(AbstractTimeSeriesDataset):
    def __init__(self, dataset: AbstractTimeSeriesDataset, indices: list):
        super().__init__()
        self.dataset = dataset
        self.indices = indices

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except Exception as e:
            return self.dataset.__getattribute__(__name)

    @property
    def n_time_series(self):
        return len(self.indices)

    def get_time_series(self, i: int):
        return self.dataset[self.indices[i]]

    def _original_index(self, i: int):
        if isinstance(self.dataset, FoldedDataset):
            return self.dataset._original_index(self.indices[i])
        else:
            return self.indices[i]

    def original_indices(self):
        return [self._original_index(i) for i in range(len(self.indices))]

    def number_of_samples_of_time_series(self, i: int) -> int:
        return self[i][0].shape[0]

    def __reduce_ex__(self, __protocol) -> Union[str, Tuple[Any, ...]]:
        return (self.__class__, (self.dataset, self.indices))


class AbstractLivesDataset(AbstractTimeSeriesDataset):
    """Base class for RUL estimation dataset

    Three abstract methods must be implemented to start
    using CERULEo with your data:

    * `__getitem__(self, i) -> pd.DataFrame`: This method should return the i-th life
    * `n_time_series(self) -> int`: The property return the total number of lives present in the dataset
    * `rul_column(self) -> str`: The property should return the name of the RUL column
    """
    @abstractproperty
    def rul_column(self) -> str:
        raise NotImplementedError

    def duration(self, life: pd.DataFrame) -> float:
        return life[self.rul_column].max()




from concurrent.futures import ProcessPoolExecutor
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.tdigest import TDigest
from sklearn.pipeline import FeatureUnion, _transform_one


class PandasToNumpy(TransformerStep):
    def fit(self, X, y=None):
        return self

    def partial_fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.values


class TransformerLambda(TransformerStep):
    def __init__(self, f, name: Optional[str] = None):
        super().__init__(name)
        self.f = f

    def transform(self, X, y=None):
        return self.f(X)


class IdentityTransformerStep(TransformerStep):
    def fit(self, input_array, y=None):
        return self

    def transform(self, input_array, y=None):
        if isinstance(input_array, pd.DataFrame):
            return input_array.copy()
        else:
            return input_array * 1


class SKLearnTransformerWrapper(TransformerStep):
    def __init__(self, transformer, name: Optional[str] = None):
        super().__init__(name)
        self.transformer = transformer

    def fit(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        self.transformer.fit(X.values)
        return self

    def partial_fit(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        if hasattr(self.transformer, "partial_fit"):
            self.transformer.partial_fit(X.values)
        else:
            self.transformer.fit(X.values)
        return self

    def _column_names(self, X) -> List[str]:
        return X.columns

    def transform(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        return pd.DataFrame(
            self.transformer.transform(X), columns=self._column_names(X), index=X.index
        )


def column_names_window(columns: list, window: int) -> list:
    """

    Parameters
    ----------
    columns: list
             List of column names

    window: int
            Window size

    Return
    ------
    Column names with the format: w_{step}_{feature_name}
    """
    new_columns = []
    for w in range(1, window + 1):
        for c in columns:
            new_columns.append(f"w_{w}_{c}")
    return new_columns


def build_tdigest(tdigest, values, column):
    return column, tdigest.merge_sorted(values)


class QuantileEstimator:
    """Approximate the quantile of each feature in the dataframe
       using t-digest
    """
    def __init__(self, tdigest_size:int = 200, max_workers:int = 1, subsample:Optional[Union[int, float]] = None):
        self.tdigest_dict = None
        self.tdigest_size = tdigest_size
        self.max_workers = max_workers
        self.subsample = subsample

    def update(self, X: pd.DataFrame):
        if X.shape[0] < 2:
            return self

        columns = X.columns
        
        if self.tdigest_dict is None:
            self.tdigest_dict = {c: TDigest(self.tdigest_size) for c in columns}

        
        results = []
        with ProcessPoolExecutor(max_workers=self.max_workers) as executor:
            for i, c in enumerate(columns):

                x = X.iloc[:, i].dropna()
                if self.subsample is not None:

                    if isinstance( self.subsample, int):
                        points_to_sample = self.subsample
                    else:
                        points_to_sample = self.subsample * X.shape[0]
                    step = max(int(X.shape[0] / float(points_to_sample)), 1)
                    x = x.iloc[::step]
                x = x.sort_values().values
                t_digest = self.tdigest_dict[c]
                results.append(executor.submit(build_tdigest, t_digest, x, c))

        for r in results:
            c, tdigest = r.result()
            self.tdigest_dict[c] = tdigest
        return self

    def estimate_quantile(self, *args, **kwargs):
        return self.quantile(*args, **kwargs)
        
    def quantile(
        self, q: float, feature: Optional[str] = None
    ) -> Union[pd.Series, float]:
        """Estimate the quantile for a set of features
        
        Parameters
        ----------
        q:float
          The quantile to estimate
        feature:Optional[Str] """
        if feature is not None:
            return self.tdigest_dict[feature].estimate_quantile(q)
        else:
            return pd.Series(
                {
                    c: self.tdigest_dict[c].estimate_quantile(q)
                    for c in self.tdigest_dict.keys()
                }
            )

class QuantileComputer:
    """Approximate the quantile of each feature in the dataframe
       using t-digest
    """
    def __init__(self, subsample_rate:float = 1):
        self.values = None
        self.tdigest_size = subsample_rate

    def update(self, X: pd.DataFrame):
        if X.shape[0] < 2:
            return self

        columns = X.columns
        
        if self.values_dict is None:
            self.values = X.copy()
        else:
            self.values = pd.concat(self.values, X)

        
        return self

    def quantile(
        self, q: float, feature: Optional[str] = None
    ) -> Union[pd.Series, float]:
        if feature is not None:
            return self.values.quantile(q)
        else:
            return self.values[feature].quantile(q)


class Literal(TransformerStep):
    def __init__(self, literal, *args):
        super().__init__(*args)
        self.literal = literal 
        
    def transform(self, X):
        return self.literal


def ensure_step(step):
    if isinstance(step, TransformerStep):
        return step    
    return Literal(step)



import numpy as np
from ceruleo.graphics.results import (barplot_errors_wrt_RUL,
                                   boxplot_errors_wrt_RUL, plot_predictions,
                                   plot_predictions_grid,
                                   shadedline_plot_errors_wrt_RUL)
from ceruleo.results.results import PredictionResult


def create_predictions(name:str, number_of_lives: int) -> PredictionResult:
    y_trues = []
    y_preds = []
    for i in range(number_of_lives):
        N = np.random.randint(1500) + 200
        y_true = np.linspace(500, 0, N)
        s = np.random.rand() * 500
        y_pred = y_true + np.random.randn(N)*s
        y_trues.append(y_true)
        y_preds.append(y_pred)

    y_true = np.hstack((*y_trues,))
    y_pred = np.hstack((*y_preds,))
    return  PredictionResult(name, y_true, y_pred)

class TestGraphics:
    def test_1(self):
        y_true = np.linspace(500, 0, num=500)
        y_pred = y_true + np.random.rand(500)*15
        r = PredictionResult('Example', y_true, y_pred)
        ax = plot_predictions_grid(r, ncols=1)
        assert ax.shape == (1, 1)

        y_true = np.hstack((y_true, y_true))
        y_pred = np.hstack((y_pred, y_pred))
        r = PredictionResult('Example', y_true, y_pred)
        ax = plot_predictions_grid(r, ncols=2)
        assert ax.shape == (1, 2)
        
        y_true = np.linspace(500, 0, num=500)
        y_pred = y_true + np.random.rand(500)*15
        y_true = np.hstack((y_true, y_true))
        y_pred = np.hstack((y_pred, y_pred))
        r = PredictionResult('Example', y_true, y_pred)
        ax = plot_predictions_grid(r, ncols=1)
        assert ax.shape == (2, 1)

    def test_boxplot(self):
        results = {
            'Model A': [
                create_predictions('Model A', 5),
                create_predictions('Model A', 5)
            ],
            'Model B': [
                create_predictions('Model B', 5),
                create_predictions('Model B', 8)
            ]
        }

        ax = boxplot_errors_wrt_RUL(results, nbins=5)
        assert ax is not None

        ax = barplot_errors_wrt_RUL(results, nbins= 5)
        assert ax is not None

        ax = shadedline_plot_errors_wrt_RUL(results, nbins=5)
        assert ax is not None

        results = { 'Model A':[
                create_predictions('Model A', 5),
            ]

        }
        ax = boxplot_errors_wrt_RUL(results, nbins=5)
        assert ax is not None


        ax = barplot_errors_wrt_RUL(results, nbins= 5)
        assert ax is not None
        
        ax = shadedline_plot_errors_wrt_RUL(results, nbins=5)
        assert ax is not None

    



import numpy as np
from ceruleo.transformation.target import PicewiseRUL


class TestTargetTransformers():
    def test_PicewiseRUL(self):
        t = PicewiseRUL(max_life=26)
        d = np.vstack(
            (np.linspace(0, 30, 50),
             np.linspace(0, 30, 50)))

        d1 = t.fit_transform(d)
        assert d1.max() == 26


import itertools
import logging
from typing import List, Optional

import numpy as np
import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractTimeSeriesDataset
from ceruleo.dataset.utils import iterate_over_features
from scipy.special import kl_div
from scipy.stats import wasserstein_distance
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


def histogram_per_life(
    life: pd.DataFrame,
    feature: str,
    bins_to_use: np.ndarray,
    normalize: bool = True,
) -> List[np.ndarray]:

    try:
        d = life[feature]
        h, _ = np.histogram(d, bins=bins_to_use)

        if normalize:
            h = h / np.sum(h)
            h += 1e-50
        return h
    except Exception as e:
        logger.info(f"Error {e} when computing the distribution for feature {feature}")


def compute_bins(ds: AbstractTimeSeriesDataset, feature: str, number_of_bins: int = 15):
    min_value = ds.get_features_of_life(0)[feature].min()
    max_value = ds.get_features_of_life(0)[feature].max()

    for life in iterate_over_features(ds):
        min_value = min(np.min(life[feature]), min_value)
        max_value = max(np.max(life[feature]), max_value)
    bins_to_use = np.linspace(min_value, max_value, number_of_bins + 1)
    return bins_to_use


def features_divergeces(
    ds: AbstractTimeSeriesDataset,
    number_of_bins: int = 15,
    columns: Optional[List[str]] = None,
    show_progress: bool = False,
) -> pd.DataFrame:
    """Compute the divergence between features

    Parameters:

        ds: The dataset
        number_of_bins: Number of bins
        columns: Which columns to use

    Returns:
        df: A DataFrame in which cach row contains the 
            distances between a feature of two run-to-failure cycle
            with the following columns:

                - Life 1: Run-to-failure cycle 1
                - Life 2: Run-to-failure cycle 2
                - W: Wasserstein
                - KL: KL Divergence
                - feature: The feature name

           
        
    """
    if columns is None:
        columns = ds.numeric_features()

    features_bins = {}
    iterator = tqdm(columns) if show_progress else columns

    for feature in iterator:
        features_bins[feature] = compute_bins(ds, feature, number_of_bins)

    histograms = {}
    for life in iterate_over_features(ds):
        for feature in columns:
            if feature not in histograms:
                histograms[feature] = []
            histograms[feature].append(
                histogram_per_life(life, feature, features_bins[feature])
            )
    
    df_data = []
    for feature in columns:
        data = {}
        for ((i, h1), (j, h2)) in itertools.combinations(enumerate(histograms[feature]), 2):
            kl = (np.mean(kl_div(h1, h2)) + np.mean(kl_div(h2, h1))) / 2
            wd = wasserstein_distance(h1, h2)
            df_data.append((i, j, wd, kl, feature))
    df = pd.DataFrame(df_data, columns=["Life 1", "Life 2", "W", "KL", "feature"])

    return df


def wasserstein_between_lives(
    life1: pd.DataFrame, life2: pd.DataFrame, columns: List[str], bins: int = 15
) -> pd.DataFrame:
    """Computer the Wasserstein distance between the features of two run-to-failure cycles

    Parameters:

        life1: Run-to-cycle failure
        life2: Run-to-cycle failure
        columns: List of features to consider
        bins: Number of bins to make the distribution

    Returns:

        w: A dataframe with one column called Wasserstein

    """

    def dropna_values(x):
        return x[~np.isnan(x)]

    data = []
    for c in columns:
        v1 = dropna_values(life1[c].values)
        v2 = dropna_values(life2[c].values)
        n_points = min(len(v1), len(v2))
        v1 = np.random.choice(v1, n_points)
        v2 = np.random.choice(v2, n_points)
        _, b = np.histogram(np.hstack((v1, v2)), bins=bins)

        h1, _ = np.histogram(v1, bins=b)
        h2, _ = np.histogram(v2, bins=b)

        # h1 = h1 / (np.nansum(h1)+ 0.0001)
        # h2 = h2 / (np.nansum(h2) + 0.0001)

    return pd.DataFrame(data, columns=["Wasserstein"], index=columns)



import itertools
import logging
from typing import Dict, List, Optional

import emd
import mmh3
import numpy as np
import pandas as pd
from numpy.lib.arraysetops import isin
from pandas.core.window.expanding import Expanding
from pyts.transformation import ROCKET as pyts_ROCKET
from functools import reduce
import pywt

# try:
#    from temporis.transformation.features.hurst import hurst_exponent
# except:
#    pass
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.rolling_windows import apply_rolling_data
from ceruleo.transformation.functional.transformers import Transformer
from ceruleo.transformation.utils import SKLearnTransformerWrapper

logger = logging.getLogger(__name__)


class TimeToPreviousBinaryValue(TransformerStep):
    """Return a column with increasing number"""

    def time_to_previous_event(self, X: pd.DataFrame, c: str):
        def min_idex(group):
            if group.iloc[0, 0] == 0:
                return np.nan
            else:
                return np.min(group.index)

        X_c_cumsum = X[[c]].cumsum()
        min_index = X_c_cumsum.groupby(c).apply(min_idex)
        X_merged = X_c_cumsum.merge(
            pd.DataFrame(min_index, columns=["start"]), left_on=c, right_index=True
        )
        return X_merged.index - X_merged["start"]

    def transform(self, X: pd.DataFrame):
        new_X = pd.DataFrame(index=X.index)
        for c in X.columns:
            new_X[f"ttp_{c}"] = self.time_to_previous_event(X, c)
        return new_X


class ColumnWiseSum(TransformerStep):
    """
    Compute the column-wise sum each column
    """

    def __init__(self, column_name: str, name: Optional[str] = None):
        super().__init__(name=name)
        self.column_name = column_name

    def transform(self, X: pd.DataFrame):
        return pd.DataFrame(X.sum(axis=1), columns=[self.column_name])


class SampleNumber(TransformerStep):
    """Return a column with increasing number"""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        df = pd.DataFrame(index=X.index)
        df["sample_number"] = list(range(X.shape[0]))
        return df


class OneHotCategorical(TransformerStep):
    """Compute a one-hot encoding for a given feature

    Parameters:
        feature: Feature name from which compute the one-hot encoding
        name: Step name, by default None
    """

    def __init__(
        self,
        *,
        feature: Optional[str] = None,
        categories: Optional[List[any]] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.feature = feature
        self.categories = categories
        self.fixed_categories = True
        if self.categories is None:
            self.categories = set()
            self.fixed_categories = False
        self.encoder = None

    def partial_fit(self, X: pd.DataFrame, y=None):
        if self.fixed_categories:
            return self
        if self.feature is None:
            self.feature = X.columns[0]
        self.categories.update(set(X[self.feature].unique()))
        return self

    def fit(self, X: pd.DataFrame, y=None):
        if self.fixed_categories:
            return self
        if self.feature is None:
            self.feature = X.columns[0]
        self.categories.update(set(X[self.feature].unique()))
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        categories = sorted(list([c for c in self.categories if c is not None]))
        d = pd.Categorical(X[self.feature], categories=categories)

        df = pd.get_dummies(d)
        df.index = X.index
        return df


class HashingEncodingCategorical(TransformerStep):
    """Compute a simple numerical encoding for a given feature

    Parameters:
        nbins: Number of bins after the hash
        feature: Feature name from which compute the simple encoding
        name: Step name
    """

    def __init__(
        self, *, nbins: int, feature: Optional[str] = None, name: Optional[str] = None
    ):
        super().__init__(name=name)
        self.nbins = nbins
        self.feature = feature
        self.categories = set()
        self.encoder = None

    def transform(self, X, y=None):
        """Return a new DataFrame with the feature  encoded with integer numbers

        Parameters
        ----------
        X : pd.DataFrame
            The input life
        y : [type], optional


        Returns
        -------
        pd.DataFrame
            A new dataframe with the same index as the input
            with 1 column
        """

        def hash(x):
            if isinstance(x, int):
                x = x.to_bytes((x.bit_length() + 7) // 8, "little")
            return (mmh3.hash(x) & 0xFFFFFFFF) % self.nbins

        if self.feature is None:
            self.feature = X.columns[0]
        X_new = pd.DataFrame(index=X.index)
        X_new["encoding"] = X[self.feature].map(hash)
        return X_new


class SimpleEncodingCategorical(TransformerStep):
    """Compute a simple numerical encoding for a given feature

    Parameters
    ----------
    feature : str
        Feature name from which compute the simple encoding
    name : Optional[str], optional
        Step name, by default None
    """

    def __init__(self, *, feature: Optional[str] = None, name: Optional[str] = None):
        super().__init__(name=name)
        self.feature = feature
        self.categories = set()
        self.encoder = None

    def partial_fit(self, X: pd.DataFrame, y=None):
        """Compute incrementally the set of possible categories

        Parameters
        ----------
        X : pd.DataFrame
            The input life

        Returns
        -------
        SimpleEncodingCategorical
            self
        """
        if self.feature is None:
            self.feature = X.columns[0]

        self.categories.update(set(X[self.feature].unique()))

        return self

    def fit(self, X: pd.DataFrame, y=None):
        """Compute the set of possible categories

        Parameters
        ----------
        X : pd.DataFrame
            The input life


        Returns
        -------
        OneHotCategorical
            self
        """
        if self.feature is None:
            self.feature = X.columns[0]

        self.categories.update(set(X[self.feature].unique()))
        return self

    def transform(self, X, y=None):
        """Return a new DataFrame with the feature  encoded with integer numbers

        Parameters
        ----------
        X : pd.DataFrame
            The input life
        y : [type], optional


        Returns
        -------
        pd.DataFrame
            A new dataframe with the same index as the input
            with 1 column
        """
        categories = sorted(list([c for c in self.categories if c is not None]))
        d = pd.Categorical(X[self.feature], categories=categories)
        return pd.DataFrame({"encoding": d.codes}, index=X.index)




def rolling_kurtosis(s: pd.Series, window, min_periods):
    return s.rolling(window, min_periods=min_periods).kurt(skipna=True)


class LifeStatistics(TransformerStep):
    """Compute diverse number of features for each life.

    Returns a 1 row with the statistics computed for every feature


    The possible features are:

    - Kurtosis
    - Skewness
    - Max
    - Min
    - Std
    - Peak
    - Impulse
    - Clearance
    - RMS
    - Shape
    - Crest
    - Hurst


    Parameters
    ----------
    to_compute : List[str], optional
        List of the features to compute, by default None
        Valid values are:
        'kurtosis', 'skewness', 'max', 'min', 'std', 'peak', 'impulse',
        'clearance', 'rms', 'shape', 'crest', 'hurst'
    name : Optional[str], optional
        Name of the step, by default None

    """

    def __init__(
        self, *,to_compute: Optional[List[str]] = None, name: Optional[str] = None
    ):

        super().__init__(name=name)
        valid_stats = [
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
        ]
        if to_compute is None:
            self.to_compute = valid_stats
        else:
            for f in to_compute:
                if f not in valid_stats:
                    raise ValueError(
                        f"Invalid feature to compute {f}. Valids are {valid_stats}"
                    )
            self.to_compute = to_compute

    def partial_fit(self, X, y=None):
        return self

    def fit(self, X, y=None):
        return self

    def _kurtosis(self, s: pd.Series):
        return s.kurt(skipna=True)

    def _skewness(self, s: pd.Series):
        return s.skew(skipna=True)

    def _max(self, s: pd.Series):
        return s.max(skipna=True)

    def _min(self, s: pd.Series):
        return s.min(skipna=True)

    def _std(self, s: pd.Series):
        return s.std(skipna=True)

    def _peak(self, s: pd.Series):
        return s.max(skipna=True) - s.min(skipna=True)

    def _impulse(self, s: pd.Series):
        m = s.abs().mean()
        if m > 0:
            return self._peak(s) / m
        else:
            return 0

    def _clearance(self, s: pd.Series):
        m = s.abs().pow(1.0 / 2).mean()
        if m > 0:
            return (self._peak(s) / m) ** 2
        else:
            return 0

    def _rms(self, s: pd.Series):
        return np.sqrt(s.pow(2).mean(skipna=True))

    def _shape(self, s: pd.Series):
        m = s.abs().mean(skipna=True)
        if m > 0:
            return self._rms(s) / m
        else:
            return 0

    def _crest(self, s: pd.Series):
        m = self._rms(s)
        if m > 0:
            return self._peak(s) / m
        else:
            return 0

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Compute features from the given life

        Parameters
        ----------
        X : pd.DataFrame
            Input life

        Returns
        -------
        pd.DataFrame
            A new DataFrame with one row with n columns.
            Let m be the number of features of the life and
            f the len(to_compute) ten  where n = m x f,
        """
        X_new = pd.DataFrame(index=[0])
        for c in X.columns:
            for stats in self.to_compute:
                X_new[f"{c}_{stats}"] = getattr(self, f"_{stats}")(X[c])
        return X_new


class RollingStatistics(TransformerStep):
    """Compute diverse number of features using an rolling window.

    For each feature present in the life a number of feature will be computed for each time stamp

    The possible features are:

    Time domain:

    - Kurtosis
    - Skewness
    - Max
    - Min
    - Std
    - Peak
    - Impulse
    - Clearance
    - RMS
    - Shape
    - Crest

    Parameters
    ----------
    window:int
        Size of the rolling window
    min_points : int, optional
        The minimun number of points of the expanding window, by default 15
    to_compute : Optional[List[Str]], optional
        Name of features to compute
    Possible values are:
    'kurtosis', 'skewness', 'max', 'min', 'std', 'peak', 'impulse',
    'clearance', 'rms', 'shape', 'crest'
    name: Optiona[str]
        Name of the step, by default None

    """

    def __init__(
        self,
        *,
        window: int = 15,
        min_points=2,
        to_compute: Optional[List[str]] = None,
        specific: Optional[Dict[str, List[str]]] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.window = window
        self.min_points = min_points
        valid_stats = [
            "mean",
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
            "deviance",
            "std_atan",
            "std_acosh",
            "std_asinh",
            "energy",
        ]

        if to_compute is not None and specific is not None:
            raise ValueError('Only one of to_compute or specific should be used')
        self.specific = specific
        self.to_compute = to_compute
        if to_compute is None:
            if specific is None:
                self.to_compute = valid_stats
            else:
                self.specific = specific
        else:
            for f in to_compute:
                if f not in valid_stats:
                    raise ValueError(
                        f"Invalid feature to compute {f}. Valids are {valid_stats}"
                    )
            self.to_compute = to_compute

    def partial_fit(self, X, y=None):
        return self

    def fit(self, X, y=None):
        return self

    def _std_asinh(self, X, rolling, abs_rolling):
        return (
            X.apply(np.arcsinh).rolling(self.window, self.min_points).std(skipna=True)
        )

    def _std_acosh(self, X, rolling, abs_rolling):
        return (
            X.apply(np.arccosh).rolling(self.window, self.min_points).std(skipna=True)
        )

    def _energy(self, X, rolling, abs_rolling):
        return X.pow(2).rolling(self.window, self.min_points).sum()

    def _std_atan(self, X, rolling, abs_rolling):
        return X.apply(np.arctan).rolling(self.window, self.min_points).std(skipna=True)

    def _mean(self, X, rolling, abs_rolling):
        return rolling.mean(skipna=True)

    def _kurtosis(self, X, rolling, abs_rolling):
        return rolling.kurt(skipna=True)

    def _skewness(self, X, rolling, abs_rolling):
        return rolling.skew(skipna=True)

    def _max(self, X, rolling, abs_rolling):
        return rolling.max(skipna=True)

    def _min(self, X, rolling, abs_rolling):
        return rolling.min(skipna=True)

    def _std(self, X, rolling, abs_rolling):
        return rolling.std(skipna=True)

    def _peak(self, X, rolling, abs_rolling):
        return rolling.max(skipna=True) - rolling.min(skipna=True)

    def _impulse(self, X, rolling, abs_rolling):
        return self._peak(X, rolling, abs_rolling) / abs_rolling.mean()

    def _deviance(self, X, rolling, abs_rolling):
        return (X - rolling.mean()) / rolling.std()

    def _clearance(self, X, rolling, abs_rolling):
        return self._peak(X, rolling, abs_rolling) / X.abs().pow(1.0 / 2).rolling(
            self.window, self.min_points
        ).mean().pow(2)

    def _rms(self, X, rolling, abs_rolling):
        return (
            X.pow(2)
            .rolling(self.window, self.min_points)
            .mean(skipna=True)
            .pow(1 / 2.0)
        )

    def _shape(self, X, rolling, abs_rolling):
        return self._rms(X, rolling, abs_rolling) / abs_rolling.mean(skipna=True)

    def _crest(self, X, rolling, abs_rolling):
        return self._peak(X, rolling, abs_rolling) / self._rms(X, rolling, abs_rolling)

    def _compute_column_names(self, X:pd.DataFrame):
        columns = []
        if self.to_compute is not None:
            for stats in self.to_compute:
                for c in X.columns:
                    columns.append(f"{c}_{stats}")
        else:
            for c in self.specific.keys():
                for stats in self.specific[c]:
                    columns.append(f"{c}_{stats}")
        return columns 

    def _transform_all_features(self, X:pd.DataFrame, X_new:pd.DataFrame, rolling, abs_rolling):

        for stats in self.to_compute:
            columns_to_assign = [f"{c}_{stats}" for c in X.columns]
            out = getattr(self, f"_{stats}")(X, rolling, abs_rolling)
            X_new.loc[:, columns_to_assign] = out.values


    def _transform_specific(self,  X:pd.DataFrame, X_new:pd.DataFrame, rolling, abs_rolling):
        for c in self.specific.keys():
            for stats in self.specific[c]: 
                feature = f"{c}_{stats}"
                out = getattr(self, f"_{stats}")(X[c], rolling[c], abs_rolling[c])
                X_new.loc[:, feature] = out.values

    def transform(self, X:pd.DataFrame):
        columns = self._compute_column_names(X)
        
        X_new = pd.DataFrame(index=X.index, columns=columns)
        rolling = X.rolling(self.window, self.min_points)
        abs_rolling = X.abs().rolling(self.window, self.min_points)
        if self.to_compute is not None:
            self._transform_all_features(X, X_new, rolling, abs_rolling)
        else:
            self._transform_specific(X, X_new, rolling, abs_rolling)
        return X_new


class ExpandingStatistics(TransformerStep):
    """Compute diverse number of features using an expandign window

    For each feature present in the life a number of feature will be computed for each time stamp

    The possible features are:

    - Kurtosis
    - Skewness
    - Max
    - Min
    - Std
    - Peak
    - Impulse
    - Clearance
    - RMS
    - Shape
    - Crest
    - Hurst


    Parameters
    ----------
    min_points : int, optional
        The minimun number of points of the expanding window, by default 2
    to_compute : List[str], optional
        List of the features to compute, by default None
        Valid values are:
        'kurtosis', 'skewness', 'max', 'min', 'std', 'peak', 'impulse',
        'clearance', 'rms', 'shape', 'crest', 'hurst'
    name : Optional[str], optional
        Name of the step, by default None

    """

    def __init__(
        self, *, min_points=2, to_compute: List[str] = None, specific: Optional[Dict[str, List[str]]] = None, name: Optional[str] = None
    ):
        super().__init__(name=name)
        self.min_points = min_points
        valid_stats = [
            "kurtosis",
            "skewness",
            "max",
            "min",
            "std",
            "peak",
            "impulse",
            "clearance",
            "rms",
            "shape",
            "crest",
            "mean",
            "deviance",
            "std_atan",
            "energy",
            "std_acosh",
            "std_asinh",
        ]
        not_default = [
            'energy',
            'deviance'
        ]
        if to_compute is not None and specific is not None:
            raise ValueError('Only one of to_compute or specific should be used')
        self.specific = specific
        self.to_compute = to_compute
        if to_compute is None:
            if specific is None:
                self.to_compute = list(set(valid_stats) - set(not_default))
            else:
                self.specific = specific
        else:
            for f in to_compute:
                if f not in valid_stats:
                    raise ValueError(
                        f"Invalid feature to compute {f}. Valids are {valid_stats}"
                    )
            self.to_compute = to_compute

    def partial_fit(self, X, y=None):
        return self

    def fit(self, X, y=None):
        return self

    def _std_asinh(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.apply(np.arcsinh).expanding(self.min_points).std(skipna=True)

    def _std_acosh(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.apply(np.arccosh).expanding(self.min_points).std(skipna=True)

    def _energy(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.pow(2).expanding(self.min_points).sum(skipna=True)

    def _std_atan(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return x.apply(np.arctan).expanding(self.min_points).std(skipna=True)

    def _kurtosis(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.kurt(skipna=True)

    def _skewness(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.skew(skipna=True)

    def _max(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.max(skipna=True)

    def _min(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.min(skipna=True)

    def _std(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.std(skipna=True)

    def _peak(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.max(skipna=True) - s.min(skipna=True)

    def _impulse(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._peak(x, s, s_abs, s_abs_sqrt, s_sq) / s_abs.mean()

    def _deviance(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return (x - s.mean(skipna=True)) / (s.std(skipna=True) + 0.00000000001)

    def _clearance(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._peak(x, s, s_abs, s_abs_sqrt, s_sq) / s_abs_sqrt.mean().pow(2)

    # def _hurst(
    #    self,
    #    x: pd.Series,
    #    s: Expanding,
    #    s_abs: Expanding,
    #    s_abs_sqrt: Expanding,
    #    s_sq: Expanding,
    # ):
    #    return s.apply(lambda s: hurst_exponent(s, method="RS"))

    def _rms(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s_sq.mean(skipna=True).pow(1 / 2.0)

    def _mean(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return s.mean(skipna=True)

    def _shape(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._rms(x, s, s_abs, s_abs_sqrt, s_sq) / s_abs.mean(skipna=True)

    def _crest(
        self,
        x: pd.Series,
        s: Expanding,
        s_abs: Expanding,
        s_abs_sqrt: Expanding,
        s_sq: Expanding,
    ):
        return self._peak(x, s, s_abs, s_abs_sqrt, s_sq) / self._rms(
            x, s, s_abs, s_abs_sqrt, s_sq
        )


    def _compute_column_names(self, X:pd.DataFrame):
        columns = []
        if self.to_compute is not None:
            for stats in self.to_compute:
                for c in X.columns:
                    columns.append(f"{c}_{stats}")
        else:
            for c in self.specific.keys():
                for stats in self.specific[c]:
                    columns.append(f"{c}_{stats}")
        return columns 

    def _transform_all_features(self, X:pd.DataFrame, X_new:pd.DataFrame, expanding, s_abs, s_abs_sqrt, s_sq):

        for stats in self.to_compute:
            columns_to_assign = [f"{c}_{stats}" for c in X.columns]
            out = getattr(self, f"_{stats}")(
                    X, expanding, s_abs, s_abs_sqrt, s_sq
                )
            X_new.loc[:, columns_to_assign] = out.values


    def _transform_specific(self,  X:pd.DataFrame, X_new:pd.DataFrame, expanding, s_abs, s_abs_sqrt, s_sq):
        for c in self.specific.keys():
            for stats in self.specific[c]: 
                feature = f"{c}_{stats}"
                out = getattr(self, f"_{stats}")(
                    X[c], expanding[c], s_abs[c], s_abs_sqrt[c], s_sq[c]
                )
                X_new.loc[:, feature] = out.values

    def transform(self, X:pd.DataFrame):
        columns = self._compute_column_names(X)
        
        X_new = pd.DataFrame(index=X.index, columns=columns)
        expanding = X.expanding(self.min_points)
        s_abs = X.abs().expanding(self.min_points)
        s_abs_sqrt = X.abs().pow(1.0 / 2).expanding(self.min_points)
        s_sq = X.pow(2).expanding(self.min_points)
        if self.to_compute is not None:
            self._transform_all_features(X, X_new, expanding, s_abs, s_abs_sqrt, s_sq)
        else:
            self._transform_specific(X, X_new, expanding, s_abs, s_abs_sqrt, s_sq)
        return X_new


class Difference(TransformerStep):
    """Compute the difference between two set of features

    Example:

        X[features1] - X[features2]

    Parameters
    
        feature_set1: Feature list of the first group to substract
        feature_set2:Feature list of the second group to substract
        name: Name of the step, by default None

    """

    def __init__(
        self, *, feature_set1: list, feature_set2: list, name: Optional[str] = None
    ):
        super().__init__(name=name)
        if len(feature_set1) != len(feature_set2):
            raise ValueError(
                "Feature set 1 and feature set 2 must have the same length"
            )
        self.feature_set1 = feature_set1
        self.feature_set2 = feature_set2
        self.feature_names_computed = False

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        if not self.feature_names_computed:
            self.feature_set1 = [self.find_feature(X, c) for c in self.feature_set1]
            self.feature_set2 = [self.find_feature(X, c) for c in self.feature_set2]
            feature_names_computed = True
        new_X = X[self.feature_set1].copy()
        new_X = new_X - X[self.feature_set2].values
        return new_X


class EMD(TransformerStep):
    """Compute the empirical mode decomposition of each feature

    Parameters
    ----------
    n : int
        Number of modes to compute
    name : Optional[str], optional
        [description], by default 'EMD'
    """

    def __init__(self, *, n: int, name: Optional[str] = "EMD"):
        super().__init__(name=name)
        self.n = n

    def transform(self, X):
        new_X = pd.DataFrame(index=X.index)
        for c in X.columns:
            try:
                imf = emd.sift.sift(X[c].values, max_imfs=self.n)
                for j in range(self.n):
                    if j < imf.shape[1]:
                        new_X[f"{c}_{j}"] = imf[:, j]
                    else:
                        new_X[f"{c}_{j}"] = np.nan
            except Exception as e:
                for j in range(self.n):
                    new_X[f"{c}_{j}"] = np.nan

        return new_X


class SlidingNonOverlappingEMD(TransformerStep):
    def __init__(
        self, *, window_size: int, max_imfs: int, keep: Optional[int] = None, **kwargs
    ):
        super().__init__(*kwargs)
        self.window_size = window_size
        self.strides = window_size
        self.max_imfs = max_imfs
        if keep is None:
            keep = self.max_imfs
        assert keep <= self.max_imfs
        self.keep = keep

    def transform(self, X: pd.DataFrame):
        def _emd(values: np.ndarray):
            out = np.zeros((values.shape[0], self.keep))
            try:
                v = emd.sift.sift(values, max_imfs=self.max_imfs)
                out[:, : self.keep] = v[:, : self.keep]
            except emd.support.EMDSiftCovergeError:
                pass

            return out

        column_list = []
        for c in X.columns:
            for i in range(self.keep):
                column_list.append(f"imf_{i}_{c}")
        out = pd.DataFrame(index=X.index, columns=column_list, dtype=np.float32)
        for c in X.columns:
            emd_computed = apply_rolling_data(
                X[c].values, _emd, self.window_size, self.strides
            )

            out.loc[:, [f"imf_{i}_{c}" for i in range(self.keep)]] = emd_computed
        return out


class EMDFilter(TransformerStep):
    """Filter the signals using Empirical Mode decomposition

    Parameters
    ----------
    n: int
       Number of
    """

    def __init__(self, *, n: int, min_imf:int, max_imf:int, name: Optional[str] = "EMD"):
        super().__init__(name=name)
        self.n = n
        self.min_imf = min_imf
        self.max_imf = max_imf

    def transform(self, X):
        new_X = pd.DataFrame(index=X.index)

        for c in X.columns:
            try:
                imf = emd.sift.sift(X[c].values, max_imfs=self.n)
                new_X[c] = np.sum(imf[:, self.min_imf:self.max_imf], axis=1)
            except Exception as e:
                new_X[c] = X[c]

        return new_X


class ChangesDetector(TransformerStep):
    """Compute how many changes there are in a categorical variable
    ['a', 'a', 'b', 'c] -> [0, 0, 1, 1]
    """

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X != X.shift(axis=0)


class Interactions(TransformerStep):
    """Compute pairwise interactions between the features"""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X_new = pd.DataFrame(index=X.index)
        for c1, c2 in itertools.combinations(X.columns, 2):
            X_new[f"{c1}_{c2}"] = X[c1] * X[c2]
        return X_new


class ROCKET(SKLearnTransformerWrapper):
    def __init__(
        self,
        *,
        n_kernels=1000,
        kernel_sizes=(7, 9, 11),
        random_state=None,
        name: Optional[str] = "ROCKET",
    ):
        transformer = pyts_ROCKET(
            n_kernels=n_kernels, kernel_sizes=kernel_sizes, random_state=random_state
        )
        super().__init__(transformer=transformer, name=name)

    def _column_names(self, X):
        a = ["Filter {i}" for i in range(self.transformer.n_kernels * 2)]
        return a


class DoG(TransformerStep):
    def __init__(
        self,
        *,
        window_size: int,
        std: List[float],
        min_points: int = 1,
        center: bool = False,
        **kwargs,
    ):
        super().__init__(*kwargs)
        self.window_size = window_size
        self.std = std
        self.center = center
        self.min_points = min_points

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X1 = X.rolling(
            window=self.window_size,
            win_type="gaussian",
            center=self.center,
            min_periods=self.min_points,
        ).mean(std=self.std[0])
        X2 = X.rolling(
            window=self.window_size,
            win_type="gaussian",
            center=self.center,
            min_periods=self.min_points,
        ).mean(std=self.std[1])
        return X1 - X2


def wrcoef(X, coef_type, coeffs, wavename, level):
    N = np.array(X).size
    a, ds = coeffs[0], list(reversed(coeffs[1:]))

    if coef_type == "a":
        return pywt.upcoef("a", a, wavename, level=level)[:N]
    elif coef_type == "d":
        return pywt.upcoef("d", ds[level - 1], wavename, level=level)[:N]
    else:
        raise ValueError("Invalid coefficient type: {}".format(coef_type))


class SlidingNonOverlappingWaveletDecomposition(TransformerStep):
    """

    # TODO TEST
    X = signal
    coeffs = pywt.wavedec(X, 'db1', level=level)
    A4 = wrcoef(X, 'a', coeffs, 'db1', level)
    D4 = wrcoef(X, 'd', coeffs, 'db1', level)
    D3 = wrcoef(X, 'd', coeffs, 'db1', 3)
    D2 = wrcoef(X, 'd', coeffs, 'db1', 2)
    D1 = wrcoef(X, 'd', coeffs, 'db1', 1)
    r = A4 + D4 + D3 + D2 + D1
    assert(np.mean(r-X) < 0.00000)

    Parameters
    ----------
    TransformerStep : [type]
        [description]
    """

    def __init__(
        self, *, window_size: int, level: int, wavelet: str, keep: List[str], **kwargs
    ):
        super().__init__(*kwargs)
        self.wavelet = wavelet
        self.level = level
        self.keep = keep
        self.window_size = window_size
        self.strides = window_size

    def transform(self, X: pd.DataFrame):
        def _wavelet(values: np.ndarray):
            coeffs = pywt.wavedec(values, self.wavelet, level=self.level)
            out = np.zeros((values.shape[0], len(self.keep)))
            for i, s in enumerate(self.keep):
                part, level = s
                out[:, i] = wrcoef(
                    values, part.lower(), coeffs, self.wavelet, int(level)
                )
            return out

        column_list = []
        for c in X.columns:
            for name in self.keep:
                column_list.append(f"wavelet_{name}_{c}")
        out = pd.DataFrame(index=X.index, columns=column_list, dtype=np.float32)
        for c in X.columns:
            wv_computed = apply_rolling_data(
                X[c].values, _wavelet, self.window_size, self.strides
            )

            out.loc[:, [f"wavelet_{name}_{c}" for name in self.keep]] = wv_computed
        return out



from typing import List, Optional, Type

import matplotlib.cm as cm
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm, Normalize


def time_series_importance(
    n_features:int,
    window_size:int,
    coefficients:np.ndarray,
    column_names: List[str],
    t: Optional[float] = None,
    ax=None,
    colormap: str = "Greens",
    features_to_plot: Optional[int] = None,
    base_alpha:float =0.2,
    normalizer_cls: Type[Normalize] = Normalize
):  
    """Plot the feature importance by time-stamp and feature

    Parameters
    ----------
    n_features : int
        Total number of features used in the model
    window_size : int
        Window size used in the model
    coefficients : np.ndarray
        Coefficient array with shape (1 x n_features*window_size)
    column_names : List[str]
        Name of the columns
    t : Optional[float], optional
        [description], by default None
    ax : [type], optional
        Axis where to put the graphic, by default None
    colormap : str, optional
        Color map to use, by default "Greens"
    features_to_plot : Optional[int], optional
        Maximum number of features to plot, by default None
        If it is omitted all the features will be used.
        The features_to_plot most important features are going to be plotted.
        The importance will be computed as the sum of the timestamp  importance
        per feature
    normalizer_cls : Type[Normalize], by  default Normalize
        Color mapper class
    """
    def color_threshold(importance, t):
        if importance > t:
            color = "green"
            alpha = 1.0
        else:
            color = "black"
            alpha = 0.2
        return color, alpha

    if ax is None:
        fig, ax = plt.subplots(figsize=(17, 5))
    else:
        fig = ax.get_figure()
    im = coefficients.reshape(window_size, n_features)
    cmap = plt.get_cmap(colormap)
    
    norm = normalizer_cls(np.min(im), np.max(im), clip=True)

    importance = np.sum(im, axis=0)

    features_order = np.argsort(importance)[::-1]
    if features_to_plot is None:
        features_to_plot = len(features_order)
    features_order = features_order[:features_to_plot][::-1]
    n_selected_features = len(features_order)

    for w in range(window_size):
        
        for y, f in enumerate(features_order):

            importance = im[w, f]
            if t is not None:
                color, alpha = color_threshold(importance, 0)
            else:
                color, alpha = cmap(norm(importance)), np.clip(
                    norm(importance) + base_alpha, 0, 1
                )

            ax.scatter(w, y, color=color, alpha=alpha, marker="s", s=75)
    ax.set_yticks(list(range(n_selected_features)))
    ax.set_yticklabels(column_names[features_order])
    ax.set_xlim(-1, window_size + 0.5)
    ax.set_ylim(-1, n_selected_features + 0.5)
    cbar = fig.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap))
    disp = (np.max(im) - np.min(im)) * 0.05
    cbar.set_ticks([np.min(im), np.max(im) - disp])
    cbar.ax.set_yticklabels(["Less important", "More Important"])
    ax.set_xlabel("Time window")
    return ax



import math

import numpy as np
from ceruleo.dataset.lives_dataset import AbstractLivesDataset, FoldedDataset
from tqdm.auto import tqdm

from sklearn.base import clone
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection._split import _BaseKFold
import multiprocessing


class RULScorerWrapper:
    def __init__(self, scorer):
        from sklearn.metrics import get_scorer

        self.scorer = get_scorer(scorer)

    def __call__(self, estimator, X, y=None):
        y_true = estimator.true_values(X)
        return self.scorer(estimator, X, y_true)

class Fitter:
    def __init__(self, model, dataset, folds_genereator, fit_kwargs):
        self.model = model
        self.folds_genereator = folds_genereator
        self.dataset = dataset 
        self.fit_kwargs = fit_kwargs

    def __call__(self, params):
        i, params  = params
        model = clone(self.model)
        model.reset()
        model.set_params(**params)
        params_results = []
        for train, validation in self.folds_genereator.split(self.dataset):
            train_dataset = FoldedDataset(self.dataset, train)
            validation_dataset = FoldedDataset(self.dataset, validation)
            model.fit(train_dataset, validation_dataset, **self.fit_kwargs)
            y_pred = model.predict(validation_dataset)
            y_true = model.true_values(validation_dataset)

            params_results.append({"mse": np.mean((y_pred - y_true) ** 2)})
        return (params, params_results)


class RULGridSearchCV:
    def __init__(self, model, params: dict, folds_genereator: _BaseKFold):
        self.params = params
        self.model = model
        self.folds_genereator = folds_genereator
        self.param_list = []
        self.results = []

    def fit(self, dataset, **fit_kwargs):
        pool = multiprocessing.Pool(6)
        self.param_list, self.results = zip(
            *pool.map(
                Fitter(self.model, dataset, self.folds_genereator, fit_kwargs),
                enumerate(list(ParameterGrid(self.params))),
            )
        )


class RULGridSearchPredefinedValidation:
    def __init__(self, model, params: dict):
        self.params = params
        self.model = model
        self.param_list = []
        self.results = []

    def fit(self, dataset_train, dataset_validation, **fit_kwargs):
        self.param_list = []
        self.results = []

        for p in tqdm(list(ParameterGrid(self.params))):
            model = clone(self.model)
            model.reset()
            model.set_params(**p)
            r = model.fit(dataset_train, dataset_validation, **fit_kwargs)
            self.param_list.append(p)
            self.results.append([r])


class GeneticAlgorithmFeatureSelection:
    def __init__(self, fitness_fun, population_size, max_iter: int):
        self.population_size = population_size
        self.fitness_fun = fitness_fun
        self.max_iter = max_iter

    def init_population(self, number_of_features):
        return (
            np.array(
                [
                    [math.ceil(e) for e in pop]
                    for pop in (
                        np.random.rand(self.population_size, number_of_features) - 0.5
                    )
                ]
            ),
            np.zeros((2, number_of_features)) - 1,
        )

    def single_point_crossover(self, population):
        r, c, n = (
            population.shape[0],
            population.shape[1],
            np.random.randint(1, population.shape[1]),
        )
        for i in range(0, r, 2):
            population[i], population[i + 1] = (
                np.append(population[i][0:n], population[i + 1][n:c]),
                np.append(population[i + 1][0:n], population[i][n:c]),
            )
        return population

    def flip_mutation(self, population):
        return population.max() - population

    def random_selection(self, population):
        r = population.shape[0]
        new_population = population.copy()
        for i in range(r):
            new_population[i] = population[np.random.randint(0, r)]
        return new_population

    def get_fitness(self, train_dataset, val_dataset, feature_list, population):
        fitness = []
        for i in range(population.shape[0]):
            # columns = [feature_list[j]
            #           for j in range(population.shape[1]) if population[i, j] == 1]
            fitness.append(self.fitness_fun(train_dataset, val_dataset, feature_list))
        return fitness

    def run(
        self,
        train_dataset,
        validation_dataset,
        feature_list,
    ):

        c = len(feature_list)

        population, memory = self.init_population(c)
        # population, memory = self.replace_duplicate(population, memory)

        fitness = self.get_fitness(
            train_dataset, validation_dataset, feature_list, population
        )

        optimal_value = max(fitness)
        optimal_solution = population[np.where(fitness == optimal_value)][0]

        for i in range(self.max_iter):
            population = self.random_selection(population)
            population = self.single_point_crossover(population)
            if np.random.rand() < 0.3:
                population = self.flip_mutation(population)

            # population, memory = replace_duplicate(population, memory)

            fitness = self.get_fitness(
                train_dataset, validation_dataset, feature_list, population
            )

            if max(fitness) > optimal_value:
                optimal_value = max(fitness)
                optimal_solution = population[np.where(fitness == optimal_value)][0]

        return optimal_solution, optimal_value



from typing import List, Optional, Tuple
import matplotlib.pyplot as plt
import matplotlib.patheffects as PathEffects
import colorsys

def add_vertical_lines_annotations(
    ax,
    v: List[Tuple[float, str]],
    colors: Optional[List] = None,
    colormap: str = "Paired",
    text_background_color: str = "w",
):

    if colors is None:
        colors = plt.get_cmap(colormap).colors[-len(v):]
    assert len(v) == len(colors)
    ymin, ymax = ax.set_ylim()
    
    half_height = (ymax-ymin)/2
    lowest_position = ymin + half_height
    for i, ((value, name), color) in enumerate(zip(v, colors)):
        hls = colorsys.rgb_to_hls(*color)
        if hls[1] > 0.5:
            text_background_color='#777777'
        else:
            text_background_color='#FAFAFA'
        
        ax.axvline(x=value, label=f"{name}: {value:.2f}", color=color, linewidth=2.5)
        
        txt = ax.text(x=value, y=lowest_position + (half_height* (i%5)/5), s=f"{name}: {value:.2f}", color=color)
        txt.set_path_effects(
            [PathEffects.withStroke(linewidth=4, foreground=text_background_color)]
        )

    ax.legend()


from copy import copy
from typing import Callable, List, Optional, Tuple, Union

import matplotlib
import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from ceruleo.dataset.ts_dataset import AbstractTimeSeriesDataset


def add_vertical_line(ax, v_x, label, color, line, n_lines):

    miny, maxy = ax.get_ylim()
    ax.axvline(v_x, label=label, color=color)
    txt = ax.text(
        v_x,
        miny + (maxy - miny) * (0.5 + 0.5 * (line / n_lines)),
        label,
        color=color,
        fontweight="semibold",
    )
    txt.set_path_effects([PathEffects.withStroke(linewidth=2, foreground="w")])


def durations_histogram(
    datasets: Union[AbstractTimeSeriesDataset, List[AbstractTimeSeriesDataset]],
    xlabel: str,
    label: Union[str, List[str]] = "",
    bins: int = 15,
    units: str = "m",
    vlines: Tuple[float, str] = [],
    ax:matplotlib.axes.Axes=None,
    add_mean: bool = True,
    add_median: bool = True,
    transform: Callable[[float], float] = lambda x: x,
    threshold: float = np.inf,
    color=None,
    **kwargs,
) ->  matplotlib.axes.Axes:
    """Generate an histogram from the lives durations of the dataset


    Example:

        durations_histogram(
            [train_dataset,validation_dataset],
            label=['Train','Validation'],
            xlabel='Unit Cycles',
            units='cycles',
            figsize=(17, 5));

    Parameters:

        datasets: Dataset from which take the lives durations
        xlabel: Label of the x axis
        label: Label of each dataset to use as label in the boxplot
        bins:  Number of bins to compute in the histogram
        units: Units of time of the lives. Useful to generate labels
        vlines: Vertical lines to be added to the plot

            Each element of the list should be the x position in the first element of the tuple,
            and the second elmenet of the tuple should be the label of the line
        ax: Axis where to draw the plot.
            If missing a new figure will be created
        add_mean: Whether to add a vertical line with the mean value
        add_median: whether to add a vertical line with the median value
        transform: A function to transform each duration
        threshold: Includes duration less than the threshold

    Returns:

        ax: The axis in which the histogram was created

    """
    if isinstance(datasets, list):
        assert len(datasets) == len(label)
    else:
        datasets = [datasets]
        label = [label]

    durations = []
    for ds in datasets:
        durations.append([transform(duration) for duration in ds.durations()])

    return histogram_from_durations(
        durations,
        xlabel=xlabel,
        label=label,
        bins=bins,
        units=units,
        vlines=vlines,
        ax=ax,
        add_mean=add_mean,
        add_median=add_median,
        threshold=threshold,
        color=color,
        **kwargs,
    )


def histogram_from_durations(
    durations: Union[List[float], List[List[float]]],
    xlabel: str,
    label: Union[str, List[str]] = "",
    bins: int = 15,
    units: str = "m",
    vlines: List[Tuple[float, str]] = [],
    ax=None,
    add_mean: bool = True,
    add_median: bool = True,
    threshold: float = np.inf,
    color=None,
    alpha=1.0,
    **kwargs,
) ->  matplotlib.axes.Axes:
    """Generate an histogram from the lives durations

    Parameters:

        durations: Duration of each run-to-failure cycle
        xlabel: Label of the x axis
        label: Label of each dataset to use as label in the boxplot
        bins:  Number of bins to compute in the histogram
        units: Units of time of the lives. Useful to generate labels
        vlines: Vertical lines to be added to the plot

            Each element of the list should be the x position in the first element of the tuple,
            and the second elmenet of the tuple should be the label of the line
        ax: Axis where to draw the plot.
            If missing a new figure will be created
        add_mean: Whether to add a vertical line with the mean value
        add_median: whether to add a vertical line with the median value
        transform: A function to transform each duration
        threshold: Includes duration less than the threshold

    Returns:

        ax: The axis in which the histogram was created
    """
    if ax is None:
        _, ax = plt.subplots(1, 1, **kwargs)

    if isinstance(durations[0], list):
        assert isinstance(label, list)
        assert len(durations) == len(label)
    else:
        durations = [durations]
        label = [label]

    for l, dur in zip(label, durations):
        if len(l) > 0:
            l += " "
        vlines = copy(vlines)
        if add_mean:
            vlines.append((np.mean(dur), l + "Mean"))
        if add_median:
            vlines.append((np.median(dur), l + "Median"))
        dur = [d for d in dur if d < threshold]
        ax.hist(dur, bins, color=color, alpha=alpha, label=l)

    ax.set_xlabel(xlabel)
    ax.set_ylabel("Number of run-to-failure cycles")

    colors = sns.color_palette("hls", len(vlines))
    for i, (v_x, l) in enumerate(vlines):
        label = f"{l}: {v_x:.2f} {units}"
        add_vertical_line(ax, v_x, label, colors[i], i, len(vlines))
    ax.legend()

    return ax.figure, ax


def durations_boxplot(
    datasets: Union[AbstractTimeSeriesDataset, List[AbstractTimeSeriesDataset]],
    xlabel: Union[str, List[str]],
    ylabel: str,
    ax:Optional[matplotlib.axes.Axes]=None,
    hlines: List[Tuple[float, str]] = [],
    units: str = "m",
    transform: Callable[[float], float] = lambda x: x,
    maxy: Optional[float] = None,
    **kwargs,
) ->  matplotlib.axes.Axes:
    """Generate boxplots of the lives duration

    Example:

        ax = durations_boxplot(
            [train_dataset, validation_dataset],
            xlabel=['Train', 'Validation'],
            ylabel='Unit Cycles',
            figsize=(17, 5))

    Parameters:

        datasets: Dataset from which take the lives durations
        xlabel:  Label of each dataset to use as label in the boxplot
        ylabel: Label of the y axis
        ax: Axis where to draw the plot.
            If missing a new figure will be created
        hlines: Horizontal lines to be added to the plot

            Each element of the list should be the y position in the first element of the tuple,
            and the second element of the tuple should be the label of the line
        units: Units of time of the lives. Useful to generate labels
        transform: A function to transform each duration
        maxy: Maximum y value of the plot


    Returns:

        ax: Axis where plot has been drawn
    """
    if isinstance(datasets, list):
        assert isinstance(xlabel, list)
        assert len(datasets) == len(xlabel)
    else:
        datasets = [datasets]
        xlabel = [xlabel]

    durations = []
    for ds in datasets:
        durations.append([transform(duration) for duration in ds.durations()])

    return boxplot_from_durations(
        durations,
        xlabel=xlabel,
        ylabel=ylabel,
        ax=ax,
        hlines=hlines,
        units=units,
        maxy=maxy,
        **kwargs,
    )


def boxplot_from_durations(
    durations: Union[List[float], List[List[float]]],
    xlabel: Union[str, List[str]],
    ylabel: str,
    ax=None,
    hlines: List[Tuple[float, str]] = [],
    units: str = "m",
    maxy: Optional[float] = None,
    **kwargs,
)->  matplotlib.axes.Axes:
    """Generate an histogram from a list of durations

    Parameters:

        durations: Durations of run-to-cycle-failure
        xlabel: Label of each dataset to use as label in the boxplot
        ylabel: Label of the y axis
        ax: Axes where the figure will be plotted
        hlines: Horizontal lines to add to the figure in the form
            [(y_coordinate, label)]
        units: Units of the y axis
        maxy: Maximum y-axis value

    Returns:
        ax: Axes where the figure has been drawn

    """
    if isinstance(durations[0], list):
        assert isinstance(xlabel, list)
        assert len(durations) == len(xlabel)
    else:
        durations = [durations]
        xlabel = [xlabel]

    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    ax.boxplot(durations, labels=xlabel)
    ax.set_ylabel(ylabel)
    if maxy is not None:
        miny, _ = ax.get_ylim()
        ax.set_ylim(miny, maxy)
    colors = sns.color_palette("hls", len(hlines))
    for i, (pos, label) in enumerate(hlines):
        ax.axhline(pos, label=f"{label}: {pos:.2f} {units}", color=colors[i])
    _, labels = ax.get_legend_handles_labels()
    if len(labels) > 0:
        ax.legend()

    return ax



import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from ceruleo.models.keras.dataset import KerasTrainableModel
from ceruleo.models.keras.losses import weighted_categorical_crossentropy
from ceruleo.models.keras.weibull import WeibullLayer
from sklearn.base import BaseEstimator, TransformerMixin
from tcn import TCN
from tensorflow.keras import Model
from tensorflow.keras.layers import (Attention, BatchNormalization, Dense,
                                     Dropout, Input)

tfd = tfp.distributions


class VariationalWeibull(WeibullLayer):
    def __init__(self, hidden):
        super(VariationalWeibull, self).__init__(
            return_params=True, name='')
        self.x1 = Dense(hidden,
                        activation='relu')

        self.scale_uniform = Dense(1, activation='relu')

        self.x2 = Dense(hidden,
                        activation='relu')

        self.k1 = Dense(1, activation='relu')
        self.k2 = Dense(1, activation='relu')

        self.uniform_sampler = tfd.Sample(
            tfd.Uniform(),
            sample_shape=(1))

    def call(self, input):

        lambda_ = self.x1(input)
        lambda_ = tf.math.exp(self.scale_uniform(lambda_))
        uniform = self.uniform_sampler.sample(tf.shape(input)[0])
        lambda_ = uniform*lambda_

        k = self.x2(input)
        k1 = self.k1(k)
        k2 = self.k2(k)
        uniform1 = self.uniform_sampler.sample(tf.shape(input)[0])

        k = tf.nn.softplus(
            k1*tf.math.pow((-1 * tf.math.log(1-uniform1)), k2) + 1)

        return self._result(lambda_, k)


class RawAndBinClasses(BaseEstimator, TransformerMixin):
    """
        A target transformer that outputs
        the RUL + nbins binary vectors
    """

    def __init__(self, nbins):
        self.nbins = nbins

    def fit(self, X, y=None):
        self.max_RUL = int(X.max())
        self.value_ranges = np.linspace(0, self.max_RUL, num=self.nbins+1)
        return self

    def transform(self, X):
        v = X
        classes = []
        for j in range(len(self.value_ranges)-1):
            lower = self.value_ranges[j]
            upper = self.value_ranges[j+1]
            classes.append(((v >= lower) & (v < upper)))
        v = np.vstack((v, *classes)).T
        return v


class SoftmaxRegression(KerasTrainableModel):
    """
    The network contains stacked layers of 1-dimensional convolutional layers
    followed by max poolings

    Parameters
    ----------
    self: list of tuples (filters: int, kernel_size: int)
          Each element of the list is a layer of the network. The first element of the tuple contaings
          the number of filters, the second one, the kernel size.
    """

    def __init__(self, raw_and_bins, alpha, window, batch_size, step, transformer, shuffle, models_path,
                 patience=4, cache_size=30, output_size=3, padding='same'):
        super(SoftmaxRegression, self).__init__(window,
                                                batch_size,
                                                step,
                                                transformer,
                                                shuffle,
                                                models_path,
                                                patience=patience,
                                                output_size=output_size,
                                                cache_size=30,
                                                callbacks=[])
        if raw_and_bins is not None:
            self.raw_and_bins = raw_and_bins
            weights = [1 for _ in range(self.raw_and_bins.nbins)]
            self.wcc = weighted_categorical_crossentropy(weights)
            self.output_size = self.raw_and_bins.nbins
        else:
            self.output_size = 1
        self.alpha = alpha

    def _generate_batcher(self, train_batcher, val_batcher):
        n_features = self.transformer.n_features

        def gen_train():
            for X, y in train_batcher:
                yield X, y

        def gen_val():
            for X, y in val_batcher:
                yield X, y

        a = tf.data.Dataset.from_generator(
            gen_train, (tf.float32, tf.float32),
            (tf.TensorShape([None, self.window, n_features]),
             tf.TensorShape([None, self.output_size])))
        b = tf.data.Dataset.from_generator(
            gen_val, (tf.float32, tf.float32),
            (tf.TensorShape([None, self.window, n_features]),
             tf.TensorShape([None, self.output_size])))
        return a, b

    def _loss(self, y_true, y_pred):
        # cross entropy loss
        bin_true = y_true[:, 1:]
        cont_true = y_true[:, 0]

        # y_pred_rul = y_pred[:, 0]
        # y_pred_bins = y_pred[:, 1:]
        y_pred_bins = y_pred
        cls_loss = self.wcc(bin_true, y_pred_bins)
        # MSE loss
        idx_tensor = self.raw_and_bins.value_ranges[:-1]
        pred_cont = tf.reduce_sum(y_pred_bins * idx_tensor, 1)
        # pred_cont = tf.keras.backend.argmax(y_pred, axis=1)
        rmse_loss_softmax = tf.losses.mean_squared_error(cont_true, pred_cont)

        # mse_loss = tf.losses.mean_squared_error(cont_true, y_pred_rul)
        # Total loss
        total_loss = (cls_loss +
                      self.alpha * rmse_loss_softmax
                      )
        return total_loss

    def mse_softmax(self, y_true, y_pred):
        # cross entropy loss
        # bin_true = y_true[:, 1:]
        cont_true = y_true[:, 0]

        # y_pred_rul = y_pred[:, 0]
        # y_pred_bins = y_pred[:, 1:]
        y_pred_bins = y_pred
        idx_tensor = self.raw_and_bins.value_ranges[:-1]
        pred_cont = tf.reduce_sum(y_pred_bins * idx_tensor, 1)
        # pred_cont = tf.keras.backend.argmax(y_pred, axis=1)
        return tf.sqrt(tf.losses.mean_squared_error(cont_true, pred_cont))

    def mse_rul(self,  y_true, y_pred):
        # cross entropy loss
        cont_true = y_true[:, 0]
        y_pred_rul = y_pred[:, 0]
        return tf.losses.mean_squared_error(cont_true, y_pred_rul)

    def compile(self):
        self.compiled = True
        self.model.compile(loss='mse',

                           optimizer=tf.keras.optimizers.Adam(lr=0.0001))

    def build_model(self):

        # function to split the input in multiple outputs
        def splitter(x):
            return [x[:, :, i:i+1] for i in range(n_features)]

        n_features = self.transformer.n_features

        i = Input(shape=(self.window, n_features))

        m = TCN(use_skip_connections=True,
                use_batch_norm=True,
                return_sequences=True,
                dropout_rate=0.1)(i)
        m = Attention(64, self.window-1)(m)
        m = Dense(100, activation='relu')(m)
        m = Dropout(0.5)(m)
        m = BatchNormalization()(m)
        proba = Dense(150, activation='relu')(m)
        proba = BatchNormalization()(proba)
        proba = Dropout(0.1)(proba)
        proba = Dense(1, activation='linear')(proba)

        return Model(inputs=i, outputs=proba)

    @property
    def name(self):
        return 'ConvolutionalSimple'

    def get_params(self, deep=False):
        params = super().get_params()
        params.update({
        })
        return params



from typing import Optional, Union

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.tdigest import TDigest
from ceruleo.transformation.utils import QuantileComputer, QuantileEstimator


class RobustMinMaxScaler(TransformerStep):
    """Scale features using statistics that are robust to outliers.

    This Scaler scales the data according to the quantile range
    The IQR is the range between the limits provided, by default,
    1st quartile (25th quantile) and the 3rd quartile (75th quantile).

    The quantiles are approximated using tdigest

    Parameters
    ----------
    range : tuple
        Desired range of transformed data.
    clip : bool, optional
        Set to True to clip transformed values of held-out data to provided, by default True
    lower_quantile : float, optional
        Lower limit of the quantile range to compute the scale, by default 0.25
    upper_quantile : float, optional
        Upper limit of the quantile range to compute the scale, by default 0.75
    tdigest_size : Optional[int], optional
        Size of the t-digest structure, by default 100
    name : Optional[str], optional
        Name of the step, by default None
    """

    def __init__(
        self,
        *,
        range: tuple,
        clip: bool = True,
        lower_quantile: float = 0.25,
        upper_quantile: float = 0.75,
        max_workers: int = 1,
        subsample: Optional[Union[int, float]] = None,
        name: Optional[str] = None,
        prefer_partial_fit:bool = False
    ):

        super().__init__(name=name, prefer_partial_fit=prefer_partial_fit)
        self.range = range
        self.Q1 = None
        self.Q3 = None
        self.clip = clip
        self.quantile_estimator = QuantileEstimator(
            tdigest_size=50, subsample=subsample, max_workers=max_workers
        )
        self.lower_quantile = lower_quantile
        self.upper_quantile = upper_quantile

    def _compute_quantiles(self):
        self.Q1 = self.quantile_estimator.quantile(self.lower_quantile)
        self.Q3 = self.quantile_estimator.quantile(self.upper_quantile)

        self.IQR = self.Q3 - self.Q1

        self.valid_mask = self.IQR.abs() > 0.000000000001


    def partial_fit(self, df: pd.DataFrame, y=None):
        self.quantile_estimator.update(df)
        return self

    def fit(self, df: pd.DataFrame, y=None):
        self.quantile_estimator.update(df)
        self._compute_quantiles()
        return self

    def transform(self, X: pd.DataFrame):
        if self.Q1 is None:
            self._compute_quantiles()
            

        new_X = X.copy()
        X_std = (X.loc[:, self.valid_mask] - self.Q1[self.valid_mask]) / (
            self.IQR[self.valid_mask]
        )
        new_X.loc[:, self.valid_mask] = (
            X_std * (self.range[1] - self.range[0]) + self.range[0]
        )
        new_X.loc[:, ~self.valid_mask] = 0
        if self.clip:
            return new_X.clip(self.range[0], self.range[1])
        else:
            return new_X


class MinMaxScaler(TransformerStep):
    """Transform features by scaling each feature to a given range.

    This transformer scales and translates each feature individually
    such that it is in the given range on the training set.

    Parameters
    ----------
    range : tuple
        Desired range of transformed data.
    clip : bool, optional
        Set to True to clip transformed values of held-out data to provided, by default True
    name : Optional[str], optional
         Name of the step, by default None
    """

    def __init__(
        self,
        *,
        range: tuple,
        clip: bool = True,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.range = range
        self.min = range[0]
        self.max = range[1]
        self.data_min = None
        self.data_max = None
        self.clip = clip

    def partial_fit(self, df, y=None):
        partial_data_min = df.min(skipna=True)
        partial_data_max = df.max(skipna=True)
        if self.data_min is None:
            self.data_min = partial_data_min
            self.data_max = partial_data_max
        else:
            self.data_min = pd.concat([self.data_min, partial_data_min], axis=1).min(
                axis=1, skipna=True
            )
            self.data_max = pd.concat([self.data_max, partial_data_max], axis=1).max(
                axis=1, skipna=True
            )
        return self

    def fit(self, df, y=None):
        self.data_min = df.min(skipna=True)
        self.data_max = df.max(skipna=True)

        return self

    def transform(self, X):
        try:
            divisor = self.data_max - self.data_min
            
            mask = np.abs((divisor)) > 1e-25
            X = X.copy()
            X.loc[:, mask] = (
                (X.loc[:, mask] - self.data_min[mask])
                / divisor[mask]
                * (self.max - self.min)
            ) + self.min
            X.loc[:, ~mask] = 0
        except:
            raise
        if self.clip:
            X.clip(lower=self.min, upper=self.max, inplace=True)
        return X

    def description(self):
        data = super().description()
        return (data, {"Min": self.data_min, "Max": self.data_max})


class StandardScaler(TransformerStep):
    """Standardize features by removing the mean and scaling to unit variance.

    Parameters
    ----------
    name : Optional[str], optional
        Name of the step, by default None

    """

    def __init__(self, *, name: Optional[str] = None):
        super().__init__(name=name)
        self.std = None
        self.mean = None

    def partial_fit(self, df, y=None):
        if df.shape[0] < 15:
            return self
        partial_data_mean = df.mean()
        partial_data_std = df.std()
        if self.mean is None:
            self.mean = partial_data_mean
            self.std = partial_data_std
        else:
            self.mean = pd.concat([self.mean, partial_data_mean], axis=1).mean(axis=1)
            self.std = pd.concat([self.std, partial_data_std], axis=1).mean(axis=1)
        return self

    def fit(self, df, y=None):
        self.mean = df.mean()
        self.std = df.std()
        return self

    def transform(self, X):

        return (X - self.mean) / (self.std)


class RobustStandardScaler(TransformerStep):
    """Scale features using statistics that are robust to outliers."""

    def __init__(self, *, quantile_range=(0.25, 0.75), prefer_partial_fit:bool = False, **kwargs):
        super().__init__( **kwargs,prefer_partial_fit=prefer_partial_fit)
        self.quantile_range = quantile_range
        self.quantile_estimator = QuantileComputer()
        self.IQR = None
        self.median = None

    def fit(self, X: pd.DataFrame, y=None):
        """Compute the mean of the dataset

        Parameters
        ----------
        X : pd.DataFrame
            the input dataset


        Returns
        -------
        MeanCentering
            self
        """
        Q1 = X.quantile(self.quantile_range[0])
        Q3 = X.quantile(self.quantile_range[1])
        self.IQR = Q3 - Q1
        self.median = X.median()

    def partial_fit(self, X: pd.DataFrame, y=None):
        """Compute incrementally the mean of the dataset

        Parameters
        ----------
        X : pd.DataFrame
            the input life

        Returns
        -------
        MeanCentering
            self
        """
        if X.shape[0] < 2:
            return self

        self.quantile_estimator.update(X)

        return self

    def _compute_quantiles(self):
        self.Q1 = self.quantile_estimator.quantile(self.quantile_range[0])
        self.Q3 = self.quantile_estimator.quantile(self.quantile_range[1])

        self.IQR = self.Q3 - self.Q1

        self.median = self.quantile_estimator.quantile(0.5)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Center the input life

        Parameters
        ----------
        X : pd.DataFrame
            The input life

        Returns
        -------
        pd.DataFrame
            A new DataFrame with the same index as the input with the
            data centered with respect to the mean of the fiited dataset
        """
        if self.IQR is None:
            self._compute_quantiles()
        return (X - self.median) / self.IQR


class ScaleInvRUL(TransformerStep):
    """
    Scale binary columns according the inverse of the RUL.
    Usually this will be used before a CumSum transformation

    Parameters
    ----------
    rul_column: str
                Column with the RUL
    """

    def __init__(self, *,rul_column: str, name: Optional[str] = None):
        super().__init__(name=name)
        self.RUL_list_per_column = {}
        self.penalty = {}
        self.rul_column_in = rul_column
        self.rul_column = None

    def partial_fit(self, X: pd.DataFrame):
        if self.rul_column is None:
            self.rul_column = self.column_name(X, self.rul_column_in)
        columns = [c for c in X.columns if c != self.rul_column]
        for c in columns:
            mask = X[X[c] > 0].index
            if len(mask) > 0:
                RUL_list = self.RUL_list_per_column.setdefault(c, [])
                RUL_list.extend(
                    (
                        1
                        + (
                            X[self.rul_column].loc[mask].values
                            / X[self.rul_column].max()
                        )
                    ).tolist()
                )

        for k in self.RUL_list_per_column.keys():

            self.penalty[k] = 1 / np.median(self.RUL_list_per_column[k])

    def transform(self, X: pd.DataFrame):
        columns = [c for c in X.columns if c != self.rul_column]
        X_new = pd.DataFrame(index=X.index)
        for c in columns:
            if c in self.penalty:
                X_new[c] = X[c] * self.penalty[c]
        return X_new


class PerCategoricalMinMaxScaler(TransformerStep):
    """Performs a minmax scaler partition the data trough some categorical feature

    Usually, different execution configurations lead to different scales in the features.
    Therefore, sometimes it is useful to scale the data based on a categorical feature,
    to reflect the difference in the execution parameters.

    Parameters
    ----------
    categorical_feature: str
        The name of the categorical feature whose values are going to be used
        to split each time-series
    scaler: Optional[Union[MinMaxScaler,RobustMinMaxScaler]], default MinMaxScaler
        The scale to use when scaling the data
    scaler_params: dict
        Parameters used when constructing the scaler
    name: Optional[str]
        Name of the transformer

    """

    def __init__(
        self,
        *,
        categorical_feature: str,
        scaler: Optional[Union[MinMaxScaler, RobustMinMaxScaler]] = MinMaxScaler,
        scaler_params: dict = {},
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.categorical_feature = categorical_feature
        self.categorical_feature_name = None
        self.scaler = scaler
        self.scaler_params = scaler_params

        self.scalers = {"default": self.scaler(**self.scaler_params)}

    def partial_fit(self, X, y=None):
        if self.categorical_feature_name is None:
            self.categorical_feature_name = self.find_feature(
                X, self.categorical_feature
            )
        for category, data in X.groupby(self.categorical_feature_name):
            data = data.drop(columns=[self.categorical_feature_name])
            if category not in self.scalers:
                self.scalers[category] = self.scaler(**self.scaler_params)
            self.scalers[category].partial_fit(data)
            self.scalers["default"].partial_fit(data)

    def transform(self, X: pd.DataFrame):

        X_new = X.drop(columns=[self.categorical_feature_name])

        for category, data in X.groupby(self.categorical_feature_name):

            data = data.drop(columns=[self.categorical_feature_name])
            scaler = (
                self.scalers[category]
                if category in self.scalers
                else self.scalers["default"]
            )  # Use a defaultdict
            X_new.loc[data.index, :] = scaler.transform(data)
        return X_new



from collections import defaultdict
from typing import Dict, List, Optional, Union

import antropy as ant
import numpy as np
import pandas as pd

from pyparsing import col
from scipy.stats import spearmanr
from ceruleo.dataset.transformed import TransformedDataset
from tqdm.auto import tqdm
from sklearn.feature_selection import mutual_info_regression
from ceruleo.dataset.ts_dataset import AbstractLivesDataset
from ceruleo.dataset.utils import iterate_over_features_and_target
from uncertainties import ufloat


def entropy(s: np.ndarray)-> float:
    """Approximate entropy

    The approximate entropy quantifies the amount of regularity and the unpredictability of fluctuations over time-series data.

    Parameters:
    
        s: A single feature

    Returns:

        ae: Approximate entropy of feature s
    """
    return ant.app_entropy(s)


def correlation(s: np.ndarray, y:Optional[np.ndarray]=None):
    """Correlation of the feature with the target

    Parameters:

        s: A single feature
        y: The RUL target

    Returns:

        c: Correlation between the feature ant the RUL target
    """
    N = s.shape[0]
    if not (s[0] == s).all():
        corr = spearmanr(s, np.arange(N), nan_policy="omit")
        corr = corr.correlation
    else:
        corr = np.nan
    return corr


def autocorrelation(s: np.ndarray)-> float:
    """Autocorrelation of a feature

    Parameters:

        s: A single feature

    Returns:

        ac: Autocorrelation of the feature
    """
    diff = np.diff(s)
    return np.sum(diff ** 2) / s.shape[0]


def monotonicity(s: np.ndarray) -> float:
    """Monotonicity of a feature

    Parameters:

        s: A single feature

    Returns:
    
        f: Monotonicity of the feature
    """
    N = s.shape[0]
    diff = np.diff(s)
    return 1 / (N - 1) * np.abs(np.sum(diff > 0) - np.sum(diff < 0))


def n_unique(s: np.ndarray) -> int:
    """Number of unique values in the array

    Parameters:

        s: A single feature

    Returns:

        n: Number of unique values
    """
    return len(np.unique(s))

def null(s: np.ndarray) -> float:
    """Null proportion for a given feature

    Parameters:
        s: A feature

    Returns:
    
        n: Null proportion
    """
    return np.mean(~np.isfinite(s))

def mutual_information(x:np.ndarray, y:np.ndarray):
    """Mutual information between a feature and the target

    [Reference](Remaining Useful Life Prediction Using Ranking Mutual Information Based Monotonic Health Indicator)

    Parameters:
    
        x: A single feature
        y: RUL Target

    Returns:    
        float: Mutual information between x and y

    """
    x = x.reshape(-1, 1)
    x = np.nan_to_num(x)
    return mutual_info_regression(x, y)

metrics = {
    "std": lambda x, y: np.std(x),
    "correlation": lambda x,y: correlation(x, y),
    "autocorrelation": lambda x,y:autocorrelation(x),
    "monotonicity": lambda x,y:monotonicity(x),
    "number_of_unique_elements": lambda x,y:n_unique(x),    
    'mutual_information': mutual_information,
    'null': lambda x, y: null(x),
    'entropy': lambda x, y: entropy(x)
}




def analysis_single_time_series(
    X: np.ndarray,
    y: np.ndarray,
    column_names: List[str],
    data: Optional[Dict] = None,
    what_to_compute: List[str] = [],
) -> dict:
    """Compute the analysis for a single run-to-failure cycle

    Parameters:

        X: Features            
        y: RUL Target
        column_names: Column names of the features
        data: Initial data
        what_to_compute: Features to compute

    Returns:
        dict: Computed info
    """

    if data is None:
        data = defaultdict(lambda: defaultdict(list))
    if len(what_to_compute) == 0:
        what_to_compute = list(sorted(metrics.keys()))
    for column_index in range(len(column_names)):
        column_name = column_names[column_index]
        for what in what_to_compute:
            x_ts = np.squeeze(X.loc[:, column_name].values)

            m = metrics[what](x_ts, y)

            
            data[column_name][what].append(m)
    return data


def merge_analysis(data: dict):
    data_df = defaultdict(lambda: defaultdict(list))
    for column_name in data.keys():
        for what in data[column_name]:
            data_df[column_name][f"{what} Mean"] = ufloat(
                np.nanmean(data[column_name][what]),
                np.nanstd(data[column_name][what]),
            )
            data_df[column_name][f"{what} Max"] = np.nanmax(data[column_name][what])
            data_df[column_name][f"{what} Min"] = np.nanmin(data[column_name][what])
    return pd.DataFrame(data_df).T


def analysis(
    dataset: Union[TransformedDataset, AbstractLivesDataset],
    *,
    show_progress: bool = False,
    what_to_compute: List[str] = [],
) -> pd.DataFrame:
    """Compute analysis of numerical features

    Parameters:

        transformed_dataset: A transformed dataset with a features and target
        show_progress: Wether to show the progress when computing the features
        what_to_compute: Elements available to compute        
                        - std
                        - correlation
                        - autocorrelation,
                        - monotonicity
                        - number_of_unique_elements
                        - mutual_information
                        - null
                        - entropy

    Returns:

        df: Dataframe with the columns specified by what_to_compute
    """

    if len(what_to_compute) == 0:
        what_to_compute = list(sorted(metrics.keys()))
    data = defaultdict(lambda: defaultdict(list))
    iterator = dataset
    if show_progress:
        iterator = tqdm(iterator)

    if isinstance(dataset, TransformedDataset):
        column_names = dataset.transformer.column_names
    else:
        column_names = dataset.numeric_features()
    for X, y in iterate_over_features_and_target(dataset):
        y = np.squeeze(y)
        data = analysis_single_time_series(
            X, y, column_names, data, what_to_compute
        )
    return merge_analysis(data)


from datetime import timedelta

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import FuncFormatter


def format_relative_time(x, pos):
    return timedelta(seconds=x)


def format_xticks(ax):
    formatter = FuncFormatter(format_relative_time)
    ax.xaxis.set_major_formatter(formatter)


def format_xaxis_time(ax):
    format_xticks(ax)
    ax.set_xlabel('Relative time from the beggining')


def ewma(feat, lambda_=0.5):
    x = np.nanmean(feat)
    s = np.sqrt(lambda_ / (2-lambda_))*np.nanstd(feat)
    UCL = x + 3*s
    LCL = x - 3*s
    return (LCL, x, UCL)


def plot_ewma(df, x_feat_name, y_feat_name, lambda_=0.5, ax=None, description='', units=None):
    plot_ewma_(df[x_feat_name],
               df[y_feat_name],
               description,
               units,
               lambda_=lambda_,
               ax=ax)


def plot_ewma_(time, feat, description, units, lambda_=0.5, ax=None, color=None):
    LCL, mean, UCL = ewma(feat, lambda_=lambda_)
    if ax is None:
        _, ax = plt.subplots(figsize=(15, 4))

    ax.plot(time, feat, color='blue')
    ax.axhline(
        y=LCL,
        label=f'LCL={np.round(LCL, 2)} | $\\lambda = {lambda_}$',
        color=color)
    ax.axhline(y=mean)
    ax.axhline(
        y=UCL,
        label=f'UCL={np.round(UCL, 2)} | $\\lambda = {lambda_}$',
        color=color)
    ax.set_title('EWMA Chart ' + description)
    # format_xaxis_time(ax)
    ax.set_ylabel(description + (f' [{units}]' if units else ''))
    # format_xaxis_time(ax)
    ax.legend()

    return LCL, mean, UCL, ax



import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
from ceruleo.graphics.plots import plot_predictions
from ceruleo.iterators.iterators import WindowedDatasetIterator
from ceruleo.iterators.utils import true_values
from ceruleo.results.results import PredictionResult
from sklearn.metrics import mean_absolute_error as mae
from tensorflow.keras.callbacks import Callback

logger = logging.getLogger(__name__)


class PredictionCallback(Callback):
    """Generate a plot after each epoch with the predictions

    Parameters:

        model: The model used predict
        output_path: Path of the output image
        dataset: The dataset that want to be plotted
    """

    def __init__(
        self,
        model: tf.keras.Model,
        output_path: Path,
        dataset: tf.data.Dataset,
        units: str='',
        filename_suffix: str = "",
    ):

        super().__init__()
        self.output_path = output_path
        self.dataset = dataset
        self.pm_model = model
        self.units = units
        self.suffix = filename_suffix
        if len(filename_suffix) > 0:
            self.output_path = self.output_path.with_stem(
                filename_suffix + "_" + self.output_path.stem
            )

    def on_epoch_end(self, epoch, logs={}):
        y_pred = self.pm_model.predict(self.dataset)
        y_true = true_values(self.dataset)
        ax = plot_predictions(
            PredictionResult('Model', y_true, y_pred),
            figsize=(17, 5),
            units=self.units,
        )
        ax.legend()

        ax.figure.savefig(self.output_path, dpi=ax.figure.dpi)

        plt.close(ax.figure)


class TerminateOnNaN(Callback):
    """Callback that terminates training when a NaN loss is encountered."""

    def __init__(self, batcher):
        super().__init__()
        self.batcher = batcher

    def on_batch_end(self, batch, logs=None):
        logs = logs or {}
        loss = logs.get("loss")
        if loss is not None:
            if np.isnan(loss) or np.isinf(loss):
                logger.info("Batch %d: Invalid loss, terminating training" % (batch))
                self.model.stop_training = True
                self.batcher.stop = True
